# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0
"""Nested sampling of jaxns.

This module is adapted from
https://github.com/pyro-ppl/numpyro/raw/master/numpyro/contrib/nested_sampling.py
"""

from __future__ import annotations

import jax
import jax.numpy as jnp
from jax import random
from numpyro.handlers import reparam, seed, trace
from numpyro.infer import Predictive
from numpyro.infer.util import (
    _guess_max_plate_nesting,
    _validate_model,
    log_density,
)

from elisa.infer.samplers.util import UniformReparam


class JAXNSSampler:
    """
    (EXPERIMENTAL) A wrapper for `jaxns <https://github.com/Joshuaalbert/jaxns>`_ ,
    a nested sampling package based on JAX.

    See reference [1] for details on the meaning of each parameter.
    Please consider citing this reference if you use the nested sampler in your research.

    .. note:: To enumerate over a discrete latent variable, you can add the keyword
        `infer={"enumerate": "parallel"}` to the corresponding `sample` statement.

    .. note:: To improve the performance, please consider enabling x64 mode at the beginning
        of your NumPyro program ``numpyro.enable_x64()``.

    **References**

    1. *JAXNS: a high-performance nested sampling package based on JAX*,
       Joshua G. Albert (https://arxiv.org/abs/2012.15286)

    :param callable model: a call with NumPyro primitives
    :param dict constructor_kwargs: additional keyword arguments to construct an upstream
        :class:`jaxns.NestedSampler` instance.
    :param dict termination_kwargs: keyword arguments to terminate the sampler. Please
        refer to the upstream :meth:`jaxns.NestedSampler.__call__` method.

    **Example**

    .. doctest::

        >>> from jax import random
        >>> import jax.numpy as jnp
        >>> import numpyro
        >>> import numpyro.distributions as dist
        >>> from numpyro.contrib.nested_sampling import NestedSampler

        >>> true_coefs = jnp.array([1., 2., 3.])
        >>> data = random.normal(random.PRNGKey(0), (2000, 3))
        >>> labels = dist.Bernoulli(logits=(true_coefs * data).sum(-1)).sample(random.PRNGKey(1))
        >>>
        >>> def model(data, labels):
        ...     coefs = numpyro.sample('coefs', dist.Normal(0, 1).expand([3]))
        ...     intercept = numpyro.sample('intercept', dist.Normal(0., 10.))
        ...     return numpyro.sample('y', dist.Bernoulli(logits=(coefs * data + intercept).sum(-1)),
        ...                           obs=labels)
        >>>
        >>> ns = NestedSampler(model)
        >>> ns.run(random.PRNGKey(2), data, labels)
        >>> samples = ns.get_samples(random.PRNGKey(3), num_samples=1000)
        >>> assert jnp.mean(jnp.abs(samples['intercept'])) < 0.05
        >>> print(jnp.mean(samples['coefs'], axis=0))  # doctest: +SKIP
        [0.93661342 1.95034876 2.86123884]
    """

    def __init__(
        self,
        model,
        *,
        constructor_kwargs=None,
        termination_kwargs=None,
    ):
        from jaxns.utils import NestedSamplerResults

        self.model = model
        self.constructor_kwargs = (
            constructor_kwargs if constructor_kwargs is not None else {}
        )
        self.termination_kwargs = (
            termination_kwargs if termination_kwargs is not None else {}
        )
        self._samples = None
        self._log_weights = None
        self._results: NestedSamplerResults | None = None

    def run(self, rng_key, *args, **kwargs):
        """
        Run the nested samplers and collect weighted samples.

        :param random.PRNGKey rng_key: Random number generator key to be used for the sampling.
        :param args: The arguments needed by the `model`.
        :param kwargs: The keyword arguments needed by the `model`.
        """
        from jaxns import Model, Prior, TerminationCondition
        from jaxns.public import DefaultNestedSampler
        import tensorflow_probability.substrates.jax.distributions as tfpd

        rng_sampling, rng_predictive = random.split(rng_key)
        # reparam the model so that latent sites have Uniform(0, 1) priors
        seeded = jax.jit(seed(self.model, rng_key))
        prototype_trace = trace(seeded).get_trace(*args, **kwargs)
        param_names = [
            site['name']
            for site in prototype_trace.values()
            if site['type'] == 'sample'
            and not site['is_observed']
            and site['infer'].get('enumerate', '') != 'parallel'
        ]
        deterministics = [
            site['name']
            for site in prototype_trace.values()
            if site['type'] == 'deterministic'
        ]
        reparam_model = reparam(
            self.model, config={k: UniformReparam() for k in param_names}
        )

        # enable enumerate if needed
        has_enum = any(
            site['type'] == 'sample'
            and site['infer'].get('enumerate', '') == 'parallel'
            for site in prototype_trace.values()
        )
        if has_enum:
            from numpyro.contrib.funsor import (
                enum,
                log_density as log_density_,
            )

            max_plate_nesting = _guess_max_plate_nesting(prototype_trace)
            _validate_model(prototype_trace)
            reparam_model = enum(reparam_model, -max_plate_nesting - 1)
        else:
            log_density_ = log_density

        def log_likelihood(params):
            params = {f'u_{k}': v for k, v in params.items()}
            return log_density_(reparam_model, args, kwargs, params)[0]

        # use NestedSampler with identity prior chain
        def prior_model():
            params = {}
            for name in param_names:
                shape = prototype_trace[name]['fn'].shape()
                param = yield Prior(
                    tfpd.Uniform(low=jnp.zeros(shape), high=jnp.ones(shape)),
                    name=f'u_{name}',
                )
                params[name] = param
            return params

        model = Model(prior_model=prior_model, log_likelihood=log_likelihood)

        default_constructor_kwargs = dict(
            num_live_points=model.U_ndims * 25,
            devices=jax.devices(),
            max_samples=1e4,
        )
        default_termination_kwargs = dict(dlogZ=1e-4)
        # Fill-in missing values with defaults. This allows user to inspect what was actually used by inspecting
        # these dictionaries
        list(
            map(
                lambda item: self.constructor_kwargs.setdefault(*item),
                default_constructor_kwargs.items(),
            )
        )
        list(
            map(
                lambda item: self.termination_kwargs.setdefault(*item),
                default_termination_kwargs.items(),
            )
        )

        default_ns = DefaultNestedSampler(
            model=model,
            **self.constructor_kwargs,
        )

        # TODO: check if this is necessary
        # jit when running on single device
        if len(default_ns.nested_sampler.devices) == 1:
            run_default_ns = jax.jit(default_ns)
        else:
            run_default_ns = default_ns

        termination_reason, state = run_default_ns(
            rng_sampling,
            term_cond=TerminationCondition(**self.termination_kwargs),
        )
        results = default_ns.to_results(
            termination_reason=termination_reason, state=state
        )

        # transform base samples back to original domains
        # Here we only transform the first valid num_samples samples
        # NB: the number of weighted samples obtained from jaxns is results.num_samples
        # and only the first num_samples values of results.samples are valid.
        num_samples = results.total_num_samples
        samples = results.samples
        predictive = Predictive(
            reparam_model, samples, return_sites=param_names + deterministics
        )
        samples = predictive(rng_predictive, *args, **kwargs)
        # replace base samples in jaxns results by transformed samples
        self._results = results._replace(samples=samples)

    def get_samples(self, rng_key, num_samples):
        """
        Draws samples from the weighted samples collected from the run.

        :param random.PRNGKey rng_key: Random number generator key to be used to draw samples.
        :param int num_samples: The number of samples.
        :return: a dict of posterior samples
        """
        from jaxns import resample

        if self._results is None:
            raise RuntimeError(
                'NestedSampler.run(...) method should be called first to obtain results.'
            )
        weighted_samples, sample_weights = self.get_weighted_samples()
        return resample(
            rng_key,
            weighted_samples,
            sample_weights,
            S=num_samples,
            replace=True,
        )

    def get_weighted_samples(self):
        """
        Gets weighted samples and their corresponding log weights.
        """
        if self._results is None:
            raise RuntimeError(
                'NestedSampler.run(...) method should be called first to obtain results.'
            )

        return self._results.samples, self._results.log_dp_mean

    def print_summary(self):
        """
        Print summary of the result. This is a wrapper of :func:`jaxns.utils.summary`.
        """
        from jaxns import summary

        if self._results is None:
            raise RuntimeError(
                'NestedSampler.run(...) method should be called first to obtain results.'
            )
        summary(self._results)

    def diagnostics(self):
        """
        Plot diagnostics of the result. This is a wrapper of :func:`jaxns.plotting.plot_diagnostics`
        and :func:`jaxns.plotting.plot_cornerplot`.
        """
        from jaxns import plot_cornerplot, plot_diagnostics

        if self._results is None:
            raise RuntimeError(
                'NestedSampler.run(...) method should be called first to obtain results.'
            )
        plot_diagnostics(self._results)
        plot_cornerplot(self._results)



"""Multiplicative models."""

from __future__ import annotations

import warnings
from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
import jax
import jax.numpy as jnp
import numpy as np

from elisa.models.model import (
    AnaIntMultiplicative,
    NumIntMultiplicative,
    ParamConfig,
)

if TYPE_CHECKING:
    from elisa.util.typing import CompEval, JAXArray, NameValMapping

__all__ = [
    'Constant',
    'Edge',
    'ExpAbs',
    'ExpFac',
    'GAbs',
    'HighECut',
    'PLAbs',
    'PhAbs',
    'TBAbs',
    'WAbs',
]


class Constant(AnaIntMultiplicative):
    r"""Energy-independent multiplicative factor.

    .. math::
        M(E) = f.

    Parameters
    ----------
    f : Parameter, optional
        The multiplicative factor :math:`f`, dimensionless.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _config = (ParamConfig('f', 'f', '', 1.0, 1e-5, 1e5),)

    @staticmethod
    def integral(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        return jnp.full(egrid.size - 1, params['f'])


class Edge(NumIntMultiplicative):
    r"""Absorption edge.

    .. math::
        M(E) =
        \begin{cases}
        \exp\left[-D \bigl(\frac{E}{E_\mathrm{c}}\bigr)^3\right],
            &\text{if } E \ge E_\mathrm{c},
        \\\\
        1, &\text{otherwise.}
        \end{cases}

    Parameters
    ----------
    Ec : Parameter, optional
        The threshold energy :math:`E_\mathrm{c}`, in units of keV.
    D : Parameter, optional
        The absorption depth :math:`D` at the threshold energy, dimensionless.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('Ec', r'E_\mathrm{c}', 'keV', 7.0, 0.0, 100.0),
        ParamConfig('D', 'D', '', 1.0, 0.0, 10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        Ec = params['Ec']
        D = params['D']
        return jnp.where(
            egrid >= Ec, jnp.exp(-D * jnp.power(egrid / Ec, 3.0)), 1.0
        )


class ExpAbs(NumIntMultiplicative):
    r"""Low-energy exponential rolloff.

    .. math::
        M(E) = \exp\left(-\frac{E_\mathrm{c}}{E}\right).

    Parameters
    ----------
    Ec : Parameter, optional
        The e-folding energy :math:`E_\mathrm{c}` for the absorption,
        in units of keV.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (ParamConfig('Ec', r'E_\mathrm{c}', 'keV', 2.0, 0.0, 200.0),)

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        return jnp.exp(-params['Ec'] / egrid)


class ExpFac(NumIntMultiplicative):
    r"""Exponential modification.

    .. math::
        M(E) =
        \begin{cases}
        1 + A \exp\bigl(-\frac{f E}{E_0}\bigr), &\text{if } E \ge E_\mathrm{c},
        \\\\
        1, &\text{otherwise,}
        \end{cases}

    where :math:`E_0` is the pivot energy fixed at 1 keV.

    Parameters
    ----------
    A : Parameter, optional
        The amplitude of effect :math:`A`, dimensionless.
    f : Parameter, optional
        The exponential factor :math:`f`, dimensionless.
    Ec : Parameter, optional
        The start energy of modification :math:`E_\mathrm{c}`, in units of keV.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('A', 'A', '', 1.0, 0.0, 1e6),
        ParamConfig('f', 'f', '', 1.0, 0.0, 1e6),
        ParamConfig('Ec', r'E_\mathrm{c}', 'keV', 0.5, 0.0, 1e6),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        A = params['A']
        f = params['f']
        Ec = params['Ec']
        return jnp.where(egrid >= Ec, 1.0 + A * jnp.exp(-f * egrid), 1.0)


class GAbs(NumIntMultiplicative):
    r"""Gaussian absorption line.

    .. math::
        M(E) = \exp\left[
                    -\frac{\tau}{\sqrt{2\pi} \sigma}
                    \exp\left[
                        -\frac{\left(E - E_\mathrm{l}\right)^2}{2 \sigma^2}
                    \right]
                \right].

    The optical depth at line center is :math:`\frac{\tau}{\sqrt{2\pi}\sigma}`.

    Parameters
    ----------
    El : Parameter, optional
        The line energy :math:`E_\mathrm{l}`, in units of keV.
    sigma : Parameter, optional
        The line width :math:`\sigma`, in units of keV.
    tau : Parameter, optional
        The line depth :math:`\tau`, in units of keV.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('El', r'E_\mathrm{l}', 'keV', 1, 0.0, 1e6),
        ParamConfig('sigma', r'\sigma', 'keV', 0.01, 0.0, 20),
        ParamConfig('tau', r'\tau', '', 1.0, 0.0, 1e6),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        El = params['El']
        sigma = params['sigma']
        tau = params['tau']
        return jnp.exp(
            -tau
            / (jnp.sqrt(2 * jnp.pi) * sigma)
            * jnp.exp(-0.5 * jnp.power((egrid - El) / sigma, 2))
        )


class HighECut(NumIntMultiplicative):
    r"""High-energy cutoff.

    .. math::
        M(E) =
        \begin{cases}
        \exp\bigl(\frac{E_\mathrm{c}-E}{E_\mathrm{f}}\bigr),
            &\text{if } E \ge E_\mathrm{c},
        \\\\
        1, &\text{otherwise.}
        \end{cases}

    Parameters
    ----------
    Ec : Parameter, optional
        The cutoff energy :math:`E_\mathrm{c}`, in units of keV.
    Ef : Parameter, optional
        The e-folding energy :math:`E_\mathrm{f}`, in units of keV.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('Ec', r'E_\mathrm{c}', 'keV', 10.0, 1e-4, 1e6),
        ParamConfig('Ef', r'E_\mathrm{f}', 'keV', 15.0, 1e-4, 1e6),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        Ec = params['Ec']
        Ef = params['Ef']
        return jnp.where(egrid >= Ec, jnp.exp((Ec - egrid) / Ef), 1.0)


class PLAbs(AnaIntMultiplicative):
    r"""Absorption as a power-law in energy. Useful for things like dust.

    .. math::
        M(E) = K \left(\frac{E}{E_0}\right)^{-\alpha},

    where :math:`E_0` is the pivot energy fixed at 1 keV.

    Parameters
    ----------
    alpha : Parameter, optional
        The power law index :math:`\alpha`, dimensionless.
    K : Parameter, optional
        The coefficient :math:`K`, dimensionless.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _config = (
        ParamConfig('alpha', r'\alpha', '', 2.0, 0.0, 5.0),
        ParamConfig('K', 'K', '', 1.0, 0.0, 100.0),
    )

    @staticmethod
    def integral(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        # ignore the case of alpha = 1.0
        one_minus_alpha = 1.0 - params['alpha']
        f = params['K'] / one_minus_alpha * jnp.power(egrid, one_minus_alpha)
        return (f[1:] - f[:-1]) / (egrid[1:] - egrid[:-1])


def _make_interp(egrid, xsect):
    xp = np.log(egrid)
    fp = np.log(xsect)
    return jax.jit(
        lambda e: jnp.exp(
            jnp.interp(
                jnp.log(e), xp, fp, left='extrapolate', right='extrapolate'
            )
        )
    )


with h5py.File(Path(__file__).parent / 'tables' / 'xsect.hdf5') as f:
    _XSECT_INTERP = {
        mabs: {
            xsect: {
                abund: _make_interp(
                    f['energy'][:], f[f'{mabs}/{xsect}/{abund}'][:]
                )
                for abund in f[f'{mabs}/{xsect}'].keys()
            }
            for xsect in f[mabs].keys()
        }
        for mabs in [k for k in f.keys() if k != 'energy']
    }
del f


class PhotonAbsorption(NumIntMultiplicative):
    r"""Photon absorption model.

    .. note::
        The photon cross-sections are obtained by interpolating cross-sections
        table, which is the same as threeml does, except that if input energy
        is outside the table energy range, extrapolated value is used.
    """

    _kwargs = ('abund', 'xsect', 'method')
    _default_abund: str
    _default_xsect: str

    def __init__(
        self,
        params: dict,
        latex: str | None,
        method: str | None,
        abund: str | None,
        xsect: str | None,
    ):
        if self.__class__.__name__ == 'WAbs':
            warnings.warn(
                'The WAbs model is obsolete and is only included for '
                'comparison with historical results. The TBAbs model '
                'should be used for the ISM or PhAbs for general '
                'photoelectric absorption.',
                DeprecationWarning,
                stacklevel=4,
            )

        if abund is None:
            abund = self._default_abund
        self.abund = abund
        if xsect is None:
            xsect = self._default_xsect
        self.xsect = xsect
        super().__init__(params, latex, method)

    @property
    def eval(self) -> CompEval:
        """Get photon absorption model function."""
        if self._continuum_jit is None:
            self._continuum_jit = jax.jit(
                self.continuum, static_argnums=(2, 3, 4)
            )

        abs_model = self.__class__.__name__.lower()
        abund = self.abund
        xsect = self.xsect
        continuum = jax.jit(
            lambda egrid, params: self._continuum_jit(
                egrid, params, abs_model, abund, xsect
            )
        )
        return self._make_integral(continuum)

    @staticmethod
    def continuum(
        egrid: JAXArray,
        params: NameValMapping,
        abs_model: str,
        abund: str,
        xsect: str,
    ) -> JAXArray:
        """Photon absorption model.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : dict
            Parameter dict for the convolution model.
        abs_model : str
            Photon absorption model name.
        abund : str
            Abundance table name.
        xsect : str
            Photon cross-sections.

        Returns
        -------
        jax.Array
            The model value at `egrid`, dimensionless.
        """
        sigma = _XSECT_INTERP[abs_model][xsect][abund](egrid)
        return jnp.exp(-params['nH'] * sigma)

    @property
    def abund(self) -> str:
        """Current abundance table."""
        return self._abund

    @abund.setter
    def abund(self, abund: str):
        abund = str(abund)
        if abund not in self.abund_list():
            available = ', '.join(self.abund_list())
            raise ValueError(f'available abundance: {available}')
        self._abund = abund

    @staticmethod
    @abstractmethod
    def abund_list() -> list[str]:
        """Get available abundance list."""
        pass

    @property
    def xsect(self) -> str:
        """Current photon cross-section."""
        return self._xsect

    @xsect.setter
    def xsect(self, xsect: str | None):
        xsect = str(xsect)
        if xsect not in self.xsect_list():
            available = ', '.join(self.xsect_list())
            raise ValueError(f'available photon cross-section: {available}')
        self._xsect = xsect

    @staticmethod
    @abstractmethod
    def xsect_list() -> list[str]:
        """Get available photon cross-section list."""
        pass


class PhAbs(PhotonAbsorption):
    r"""Photoelectric absorption.

    .. math ::
        M(E) = \exp \left[ -\eta_\mathrm{H}\ \sigma(E) \right],

    where :math:`\sigma(E)` is the photo-electric cross-section, **NOT**
    including Thomson scattering.

    Parameters
    ----------
    nH : Parameter, optional
        The equivalent hydrogen column density :math:`\eta_\mathrm{H}`,
        in units of 10²² cm⁻².
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    abund : str, optional
        Abundance table to use. Available options are:

            * ``'angr'`` [1]_ (Photospheric, using Table 2)
            * ``'aspl'`` [2]_ (Photospheric, using Table 1)
            * ``'feld'`` [3]_
            * ``'aneb'`` [4]_
            * ``'grsa'`` [5]_
            * ``'wilm'`` [6]_
            * ``'lodd'`` [7]_ (Photospheric, using Table 1)
            * ``'lpgp'`` [8]_ (Photospheric, using Table 4)
            * ``'lpgs'`` [8]_ (Proto-solar, using Table 10)

        The default is ``'angr'``.
    xsect : str, optional
        Photon cross-section to use. Available options are:

            * ``'bcmc'`` [9]_ with a new He cross-section [10]_
            * ``'obcm'`` [9]_
            * ``'vern'`` [11]_

        The default is ``'vern'``.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.

    References
    ----------
    .. [1] `Anders & Grevesse 1989, Geochimica et Cosmochimica Acta, 53, 1,
            197-214 <https://doi.org/10.1016/0016-7037(89)90286-X>`__
    .. [2] `Asplund et al. 2009 ARAA, 47, 481 <https://doi.org/10.1146/annurev.astro.46.060407.145222>`__
    .. [3] `Feldman 1992, Phys. Scr. 46, 202 <https://doi.org/10.1088/0031-8949/46/3/002>`__
    .. [4] `Anders & Ebihara 1982, Geochimica et Cosmochimica Acta, 46, 11,
            2363-2380 <https://doi.org/10.1016/0016-7037(82)90208-3>`__
    .. [5] `Grevesse & Sauval 1998, Space Science Reviews, 85, 161–174 <https://doi.org/10.1023/A:1005161325181>`__
    .. [6] `Wilms et al 2000, ApJ, 542, 914 <https://doi.org/10.1086/317016>`__
    .. [7] `Lodders 2003, ApJ, 591, 1220 <https://doi.org/10.1086/375492>`__
    .. [8] `Lodders et al. 2009 <https://doi.org/10.1007/978-3-540-88055-4_34>`__
    .. [9] `Balucinska-Church & McCammon 1992, ApJ, 400, 699 <https://doi.org/10.1086/172032>`__
    .. [10] `Yan et al. 1998, ApJ, 496, 1044 <https://doi.org/10.1086/305420>`__
    .. [11] `Verner et al. 1996, ApJ, 465, 487 <https://doi.org/10.1086/177435>`__
    """

    _default_abund: str = 'angr'
    _default_xsect: str = 'vern'
    _config = (
        ParamConfig('nH', r'\eta_\mathrm{H}', '10^22 cm^-2', 1.0, 0.0, 1e6),
    )

    @staticmethod
    def abund_list() -> list[str]:
        """Get available abundance list."""
        return [
            'angr',
            'aspl',
            'feld',
            'aneb',
            'grsa',
            'wilm',
            'lodd',
            'lpgp',
            'lpgs',
        ]

    @staticmethod
    def xsect_list() -> list[str]:
        """Get available photon cross-section list."""
        return ['bcmc', 'obcm', 'vern']


class TBAbs(PhotonAbsorption):
    r"""The Tuebingen-Boulder ISM absorption model.

    This model calculates the cross-sections for X-ray absorption by the ISM as
    the sum of the cross-sections for X-ray absorption due to the gas-phase
    ISM, the grain-phase ISM, and the molecules in the ISM.

    Parameters
    ----------
    nH : Parameter, optional
        The equivalent hydrogen column density :math:`\eta_\mathrm{H}`,
        in units of 10²² cm⁻².
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    abund : str, optional
        Abundance table to use. Available options are:

            * ``'angr'`` [1]_ (Photospheric, using Table 2)
            * ``'aspl'`` [2]_ (Photospheric, using Table 1)
            * ``'feld'`` [3]_
            * ``'aneb'`` [4]_
            * ``'grsa'`` [5]_
            * ``'wilm'`` [6]_
            * ``'lodd'`` [7]_ (Photospheric, using Table 1)
            * ``'lpgp'`` [8]_ (Photospheric, using Table 4)
            * ``'lpgs'`` [8]_ (Proto-solar, using Table 10)

        The default is ``'wilm'``.
    xsect : str, optional
        Always use cross-section ``'vern'`` [9]_ as baseline.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.

    References
    ----------
    .. [1] `Anders & Grevesse 1989, Geochimica et Cosmochimica Acta, 53, 1,
            197-214 <https://doi.org/10.1016/0016-7037(89)90286-X>`__
    .. [2] `Asplund et al. 2009 ARAA, 47, 481 <https://doi.org/10.1146/annurev.astro.46.060407.145222>`__
    .. [3] `Feldman 1992, Phys. Scr. 46, 202 <https://doi.org/10.1088/0031-8949/46/3/002>`__
    .. [4] `Anders & Ebihara 1982, Geochimica et Cosmochimica Acta, 46, 11,
            2363-2380 <https://doi.org/10.1016/0016-7037(82)90208-3>`__
    .. [5] `Grevesse & Sauval 1998, Space Science Reviews, 85, 161–174 <https://doi.org/10.1023/A:1005161325181>`__
    .. [6] `Wilms et al 2000, ApJ, 542, 914 <https://doi.org/10.1086/317016>`__
    .. [7] `Lodders 2003, ApJ, 591, 1220 <https://doi.org/10.1086/375492>`__
    .. [8] `Lodders et al. 2009 <https://doi.org/10.1007/978-3-540-88055-4_34>`__
    .. [9] `Verner et al. 1996, ApJ, 465, 487 <https://doi.org/10.1086/177435>`__
    """

    _default_abund: str = 'wilm'
    _default_xsect: str = 'vern'
    _config = (
        ParamConfig('nH', r'\eta_\mathrm{H}', '10^22 cm^-2', 1.0, 0.0, 1e6),
    )

    @staticmethod
    def abund_list() -> list[str]:
        """Get available abundance list."""
        return [
            'angr',
            'aspl',
            'feld',
            'aneb',
            'grsa',
            'wilm',
            'lodd',
            'lpgp',
            'lpgs',
        ]

    @staticmethod
    def xsect_list() -> list[str]:
        return ['vern']


class WAbs(PhotonAbsorption):
    r"""A photo-electric absorption using Wisconsin cross-sections [1]_.

    .. math ::
        M(E) = \exp \left[ -\eta_\mathrm{H}\ \sigma(E) \right],

    where :math:`\sigma(E)` is the photo-electric cross-section, **NOT**
    including Thomson scattering.

    .. warning ::
        The :class:`WAbs` model is obsolete and is only included for comparison
        with historical results. The :class:`TBAbs` model should be used for
        the ISM or :class:`PhAbs` for general photoelectric absorption.

    Parameters
    ----------
    nH : Parameter, optional
        The equivalent hydrogen column density :math:`\eta_\mathrm{H}`,
        in units of 10²² cm⁻².
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    abund : str, optional
        Always use abundance table ``'aneb'`` [2]_.
    xsect : str, optional
        Always use Wisconsin cross-sections [1]_.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.

    References
    ----------
    .. [1] `Morrison & McCammon 1983, ApJ, 270, 119 <https://doi.org/10.1086/161102>`__
    .. [2] `Anders & Ebihara 1982, Geochimica et Cosmochimica Acta, 46, 11,
            2363-2380 <https://doi.org/10.1016/0016-7037(82)90208-3>`__
    """

    _default_abund: str = 'aneb'
    _default_xsect: str = 'wabs'
    _config = (
        ParamConfig('nH', r'\eta_\mathrm{H}', '10^22 cm^-2', 1.0, 0.0, 1e6),
    )

    @staticmethod
    def abund_list() -> list[str]:
        return ['aneb']

    @staticmethod
    def xsect_list() -> list[str]:
        return ['wabs']



from __future__ import annotations

import re
import warnings
from typing import TYPE_CHECKING

import numpy as np
from astropy.io import fits
from astropy.units import Unit
from scipy.sparse import coo_array

from elisa.data.base import ObservationData, ResponseData, SpectrumData
from elisa.util.misc import to_native_byteorder

if TYPE_CHECKING:
    NDArray = np.ndarray


class Data(ObservationData):
    """Handle observation data in OGIP standards [1]_ [2]_.

    Load the observation spectrum, the telescope response and the possible
    background, and handle the grouping of spectrum and response.

    Parameters
    ----------
    erange : array_like
        Energy range of interest in keV, e.g., ``erange=[(0.5, 2), (5, 200)]``.
    specfile : str
        Spectrum file path. For type II pha file, the row specifier must be
        given in the end of path, e.g., ``specfile='spec.pha2{1}'``.
    backfile : str or None, optional
        Background file path. Read from the `specfile` header if None.
        For type II pha file, the row specifier must be given in the end of
        path, e.g., ``backfile='back.pha2{1}'``.
    respfile : str or None, optional
        Response file path. Read from the `specfile` header if None.
        The path must be given if ``RESPFILE`` is undefined in the header.
    ancrfile : str or None, optional
        Ancillary response path. Read from the `specfile` header if None.

    Other Parameters
    ----------------
    name : str or None, optional
        Data name. Read from the `specfile` header if None. The name must
        be given if ``DETNAM``, ``INSTRUME`` and ``TELESCOP`` are all
        undefined in the header.
    group : str or None, optional
        Method to group spectrum and background adaptively, these options are
        available so that each channel group has:

            * ``'const'``: `scale` number channels
            * ``'min'``: total (source + background) counts >= `scale`
            * ``'sig'``: source significance >= `scale` sigma
            * ``'bmin'``: background counts >= `scale`, used to avoid bias when
              using ``wstat`` to simultaneously fit the source and background
            * ``'bsig'``: background significance >= `scale` sigma, used to
              avoid bias when using ``pgstat`` to simultaneously fit the source
              and background
            * ``'opt'``: optimal binning, see Kaastra & Bleeker (2016) [3]_
            * ``'optmin'``: optimal binning with total counts >= `scale`
            * ``'optsig'``: optimal binning with source significance >= `scale`
              sigma
            * ``'optbmin'``: optimal binning with background counts >= `scale`
            * ``'optbsig'``: optimal binning with background significance
              >= `scale` sigma

        The default is None.
    scale : float or None, optional
        Grouping scale for the method specified in `group`.
    preserve_data_group : bool, optional
        Whether to preserve the grouping flags stored in `specfile`.
        If true, will first group the data based on the grouping flags stored
        in `specfile`, then apply the grouping method specified in `group`.
        If false, will directly apply the grouping method specified in
        `group` to the data, ignoring the grouping flags stored in `specfile`.
        The default is False.
    spec_poisson : bool or None, optional
        Whether the spectrum data follows counting statistics, reading from
        the `specfile` header. This value must be set if ``POISSERR`` is
        undefined in the header.
    back_poisson : bool or None, optional
        Whether the background data follows counting statistics, reading
        from the `backfile` header. This value must be set if ``POISSERR``
        is undefined in the header.
    ignore_bad : bool, optional
        Whether to ignore channels with ``QUALITY`` being 2 or 5.
        The default is True. The possible values for spectral ``QUALITY`` are

            * ``0``: good
            * ``1``: defined bad by software
            * ``2``: defined dubious by software
            * ``5``: defined bad by user
            * ``-1``: reason for bad flag unknown

    keep_channel_info : bool, optional
        Whether to keep channel information in the label of grouped
        channel. Takes effect only if `group` is not None or spectral data
        has ``GROUPING`` defined. The default is False.
    sparse_response : bool, optional
        Whether the response matrix is sparse. The default is False.

    Notes
    -----
    Reading and applying correction to data is not yet supported.

    References
    ----------
    .. [1] `The OGIP Spectral File Format <https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007/ogip_92_007.html>`__
            and `Addendum: Changes log <https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007a/ogip_92_007a.html>`__
    .. [2] `The Calibration Requirements for Spectral Analysis (Definition of
            RMF and ARF file formats) <https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002/cal_gen_92_002.html>`__
            and `Addendum: Changes log <https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002a/cal_gen_92_002a.html>`__
    .. [3] `Kaastra & Bleeker 2016, A&A, 587, A151 <https://doi.org/10.1051/0004-6361/201527395>`__
    """

    def __init__(
        self,
        erange: list | tuple,
        specfile: str,
        backfile: str | None = None,
        respfile: str | None = None,
        ancrfile: str | None = None,
        name: str | None = None,
        group: str | None = None,
        scale: float | None = None,
        preserve_data_group: bool = False,
        spec_poisson: bool | None = None,
        back_poisson: bool | None = None,
        ignore_bad: bool = True,
        keep_channel_info: bool = False,
        sparse_response: bool = False,
        # corrfile: bool | None = None,
        # corrnorm: bool | None = None,
    ):
        try:
            spec_data = Spectrum(specfile, spec_poisson)
        except PoissonFlagNotFoundError as err:
            raise PoissonFlagNotFoundError(
                f'"POISSERR" is undefined in header of {specfile}, '
                'spec_poisson must be set manually, i.e., '
                'Data(..., spec_poisson=True/False)'
            ) from err

        # check data name
        if name:
            name = str(name)
        elif spec_data.name:
            name = spec_data.name
        else:
            raise ValueError(
                f'name must be set manually for {specfile} data, i.e., '
                "Data(..., name='NAME')"
            )

        # check ancillary response file
        if not ancrfile:
            ancrfile = spec_data.ancrfile

        # check response file
        sparse_response = bool(sparse_response)
        if respfile:
            resp_data = Response(respfile, ancrfile, sparse_response)
        elif spec_data.respfile:
            resp_data = Response(spec_data.respfile, ancrfile, sparse_response)
        else:
            raise ValueError(
                f'response file must be set manually for {specfile} data, '
                "i.e., Data(..., respfile='/path/to/rsp.fits')"
            )

        if len(spec_data.counts) != resp_data.channel_number:
            raise ValueError(
                f'specfile ({specfile}) and respfile ({respfile}) are not '
                'matched'
            )

        # check background file
        try:
            if backfile:
                back_data = Spectrum(backfile, back_poisson)
            elif spec_data.backfile:
                back_data = Spectrum(spec_data.backfile, back_poisson)
            else:
                back_data = None
        except PoissonFlagNotFoundError as err:
            raise PoissonFlagNotFoundError(
                '"POISSERR" is undefined in header of background spectrum of '
                f'{specfile}, back_poisson must be set manually, i.e., '
                'Data(..., back_poisson=True/False)'
            ) from err

        if back_data and len(back_data.counts) != resp_data.channel_number:
            raise ValueError(
                f'specfile ({specfile}) and backfile ({backfile}) are not '
                'matched'
            )

        super().__init__(
            name=name,
            erange=erange,
            spec_data=spec_data,
            resp_data=resp_data,
            back_data=back_data,
            ignore_bad=bool(ignore_bad),
            keep_channel_info=bool(keep_channel_info),
        )

        if group is not None:
            self.group(group, scale, preserve_data_group)


class Spectrum(SpectrumData):
    """Load spectrum data in OGIP standards [1]_.

    Parameters
    ----------
    specfile : str
        Spectrum file path. For type II pha file, the row specifier must be
        given at the end of path, e.g., ``specfile='spec.pha2{1}'``.
    poisson : bool or None, optional
        Whether the spectrum data follows counting statistics, reading from
        the `specfile` header. This value must be set if ``POISSERR`` is
        undefined in the header.

    References
    ----------
    .. [1] `The OGIP Spectral File Format <https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007/ogip_92_007.html>`__
            and `Addendum: Changes log <https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007a/ogip_92_007a.html>`__
    """

    def __init__(self, specfile: str, poisson: bool | None = None):
        # test if file is '/path/to/specfile{n}'
        match = re.compile(r'(.+){(.+)}').match(specfile)
        if match:
            file = match.group(1)
            type_ii = True  # spectrum file is of type II
            spec_id = int(match.group(2)) - 1  # row specifier to index
        else:
            file = specfile
            type_ii = False
            spec_id = 0

        with fits.open(file) as hdul:
            header = hdul['SPECTRUM'].header
            data = hdul['SPECTRUM'].data

        # TODO: more robust way to detect a type II data
        # check if data is type II
        if not type_ii:
            msg = 'row id must be provided for type II spectrum, i.e., '
            msg += f"'{specfile}{{N}}'"

            channel_number = len(data)
            if int(header.get('DETCHANS', channel_number)) != channel_number:
                raise ValueError(msg)

            if header.get('HDUCLAS4', '') == 'TYPE:II':
                raise ValueError(msg)

        else:
            data = data[spec_id : spec_id + 1]  # set data to the specified row

        # check if COUNTS or RATE exists
        if (
            'COUNTS' not in data.names
            and 'RATE' not in data.names
            and 'RATES' not in data.names
        ):
            raise ValueError(
                f'"COUNTS", "RATE" or "RATES" not found in {specfile}'
            )

        # get poisson flag
        poisson = header.get('POISSERR', poisson)
        if poisson is None:
            raise PoissonFlagNotFoundError(
                '"POISSERR" is undefined in header, `poisson` must be set in '
                'Spectrum(..., poisson=True/False)'
            )

        # check if STAT_ERR exists for non-Poisson spectrum
        if not poisson and 'STAT_ERR' not in data.names:
            raise ValueError(f'"STAT_ERR" not found in {specfile}')

        def get_field(field, default=None, excluded=None):
            """Get value of specified field, return default if not found."""
            if field in data.names:
                value = data[field]
                if type_ii:
                    value = value[0]
            else:
                value = header.get(field, default)

            if excluded is not None and value in excluded:
                return default
            else:
                return value

        # get exposure
        exposure = np.float64(get_field('EXPOSURE'))

        # get counts
        if 'COUNTS' in data.names:
            counts = get_field('COUNTS')
            counts = np.array(counts, dtype=np.float64, order='C')
        else:  # calculate counts using 'RATE' and 'EXPOSURE'
            if 'RATE' in data.names:
                rate = get_field('RATE')
            else:
                rate = get_field('RATES')
            rate = np.array(rate, dtype=np.float64, order='C')
            counts = rate * exposure

        # get statistical error of counts
        if poisson:
            stat_err = np.sqrt(counts)
        else:
            stat_err = get_field('STAT_ERR')
            stat_err = np.array(stat_err, dtype=np.float64, order='C')
            if 'RATE' in data.names or 'RATES' in data.names:
                stat_err *= exposure

                if 'COUNTS' in data.names:
                    warnings.warn(
                        f'"STAT_ERR" in {specfile} is assumed for "RATE"',
                        Warning,
                        stacklevel=3,
                    )

        # get fractional systematic error of counts
        sys_err = get_field('SYS_ERR', 0)
        if np.shape(sys_err) == ():
            sys_err = np.zeros(len(counts))
        else:
            sys_err = np.array(sys_err, dtype=np.float64, order='C')

        # get quality flag
        quality = get_field('QUALITY', 0)
        if np.shape(quality) == ():
            quality = np.zeros(len(counts), dtype=np.int64)
        else:
            quality = np.array(quality, dtype=np.int64, order='C')

        # get grouping flag
        grouping = get_field('GROUPING', 0)
        if np.shape(grouping) == ():
            grouping = np.ones(len(counts), np.int64)
        else:
            grouping = np.array(grouping, dtype=np.int64, order='C')

        # check data
        if poisson:
            # check if counts are integers
            diff = np.abs(counts - np.round(counts))
            if np.any(diff > 1e-8 * counts):
                warnings.warn(
                    f'Poisson spectrum {specfile} has non-integer counts, '
                    'which may lead to wrong result',
                    Warning,
                    stacklevel=3,
                )
        else:
            # check if statistical errors are positive
            if np.any(stat_err < 0.0):
                raise ValueError(
                    f'spectrum {specfile} has negative statistical errors'
                )

            if np.any(stat_err == 0.0):
                warnings.warn(
                    f'spectrum {specfile} has zero statistical errors, '
                    'which may lead to wrong result under Gaussian statistics,'
                    ' consider grouping the spectrum',
                    Warning,
                    stacklevel=3,
                )

            # check if systematic errors are non-negative
            if np.any(sys_err < 0.0):
                raise ValueError(
                    f'spectrum {specfile} has systematic errors < 0'
                )

        # total error of counts
        if not poisson:
            stat_var = np.square(stat_err)
            sys_var = np.square(sys_err * counts)
            errors = np.sqrt(stat_var + sys_var)
        else:
            if np.any(sys_err > 0.0):
                warnings.warn(
                    'systematic errors are ignored for Poisson spectrum '
                    f'{specfile}',
                    Warning,
                    stacklevel=3,
                )
            errors = stat_err

        # search name in header
        excluded_name = ('', 'none', 'unknown')
        for key in ('DETNAM', 'INSTRUME', 'TELESCOP'):
            name = str(header.get(key, ''))
            if name.lower() not in excluded_name:
                break
            else:
                name = ''
        self._name = str(name)

        # get the area scaling factor
        area_scale = get_field('AREASCAL', 1.0)
        if isinstance(area_scale, np.ndarray):
            area_scale = np.array(area_scale, dtype=np.float64, order='C')
        else:
            area_scale = float(area_scale)

        # get the background scaling factor
        back_scale = get_field('BACKSCAL', 1.0)
        if isinstance(back_scale, np.ndarray):
            back_scale = np.array(back_scale, dtype=np.float64, order='C')
        else:
            back_scale = float(back_scale)

        # get the correction scaling factor
        # self._corr_scale = np.float64(get_field('CORRSCAL', 0.0))

        self._header = header
        self._specfile = specfile
        excluded_file = ('none', 'None', 'NONE')

        # get backfile
        self._backfile = get_field('BACKFILE', '', excluded_file)

        # get respfile
        self._respfile = get_field('RESPFILE', '', excluded_file)

        # get ancrfile
        self._ancrfile = get_field('ANCRFILE', '', excluded_file)

        # get corrfile
        # self._corrfile = get_field('CORRFILE', '', excluded_file)

        super().__init__(
            counts=counts,
            errors=errors,
            poisson=poisson,
            exposure=exposure,
            quality=quality,
            grouping=grouping,
            area_scale=area_scale,
            back_scale=back_scale,
            sys_errors=sys_err,
            zero_errors_warning=False,
            non_int_warning=False,
            sys_errors_warning=False,
        )

    @property
    def name(self) -> str:
        """The name of the observation instrument."""
        return self._name

    @property
    def header(self) -> fits.Header:
        """The spectrum header."""
        return self._header

    @property
    def specfile(self) -> str:
        """The spectrum file path."""
        return self._specfile

    @property
    def backfile(self) -> str:
        """The background file path."""
        return self._backfile

    @property
    def respfile(self) -> str:
        """The response file path."""
        return self._respfile

    @property
    def ancrfile(self) -> str:
        """The ancillary response file path."""
        return self._ancrfile


class Response(ResponseData):
    """Load telescope response data in OGIP standards [1]_.

    Parameters
    ----------
    respfile : str
        Response file path.
    ancrfile : str or None, optional
        Ancillary response path. The default is None.
    sparse : bool, optional
        Whether the response matrix is sparse. The default is False.

    References
    ----------
    .. [1] `The Calibration Requirements for Spectral Analysis (Definition of
            RMF and ARF file formats) <https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002/cal_gen_92_002.html>`__
            and `Addendum: Changes log <https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002a/cal_gen_92_002a.html>`__
    """

    def __init__(
        self,
        respfile: str,
        ancrfile: str | None = None,
        sparse: bool = False,
    ):
        if ancrfile is None:
            ancrfile = ''

        self._respfile = respfile
        self._ancrfile = ancrfile

        # test if file is '/path/to/respfile{n}'
        match = re.compile(r'(.+){(.+)}').match(respfile)
        if match:  # respfile file is of type II
            file = match.group(1)
            resp_id = int(match.group(2))
        else:
            file = respfile
            resp_id = 1

        response_data = self._read_response(file, resp_id)

        arf = self._read_arf()
        if arf is not None:
            if len(arf) != response_data['response_matrix'].shape[0]:
                raise ValueError(
                    f'rmf ({respfile}) and arf ({ancrfile}) are not matched'
                )
            response_data['response_matrix'] *= arf[:, None]

        # photon_egrid, sparse_matrix = self._drop_zeros(
        #     response_data['photon_egrid'],
        #     response_data['sparse_matrix']
        # )
        # response_data['photon_egrid'] = photon_egrid
        # response_data['sparse_matrix'] = sparse_matrix

        super().__init__(**response_data, sparse=sparse)

    def _read_response(self, file: str, response_id: int) -> dict:
        respfile = self.respfile
        with fits.open(file) as response_hdul:
            if 'MATRIX' in response_hdul:
                if self.ancrfile is None:
                    warnings.warn(
                        f'{file} is probably a rmf, '
                        'ancrfile (arf) maybe needed but not provided',
                        Warning,
                    )

                ext = ('MATRIX', response_id)

            elif 'SPECRESP MATRIX' in response_hdul:
                ext = ('SPECRESP MATRIX', response_id)

            else:
                # try to find the first extension with MATRIX data
                for hdu in response_hdul:
                    if hdu.data is not None and 'MATRIX' in hdu.data.names:
                        ext = (hdu.name, response_id)
                        break
                else:
                    raise ValueError(
                        f'cannot read response matrix data from {respfile}'
                    )

            ebounds_header = response_hdul['EBOUNDS'].header
            ebounds_data = response_hdul['EBOUNDS'].data
            response_header = response_hdul[ext].header
            response_data = response_hdul[ext].data

        channel_type = ebounds_header.get(
            'CHANTYPE', response_header.get('CHANTYPE', 'PI')
        )

        emin_idx = ebounds_data.names.index('E_MIN') + 1
        emax_idx = ebounds_data.names.index('E_MAX') + 1
        emin_unit = ebounds_header.get(f'TUNIT{emin_idx}', 'keV')
        emax_unit = ebounds_header.get(f'TUNIT{emax_idx}', 'keV')
        channel_emin = ebounds_data['E_MIN'] * Unit(emin_unit).to('keV')
        channel_emax = ebounds_data['E_MAX'] * Unit(emax_unit).to('keV')

        elo_idx = response_data.names.index('ENERG_LO') + 1
        ehi_idx = response_data.names.index('ENERG_HI') + 1
        elo_unit = response_header.get(f'TUNIT{elo_idx}', 'keV')
        ehi_unit = response_header.get(f'TUNIT{ehi_idx}', 'keV')
        photon_emin = response_data['ENERG_LO'] * Unit(elo_unit).to('keV')
        photon_emax = response_data['ENERG_HI'] * Unit(ehi_unit).to('keV')
        photon_egrid = np.append(photon_emin, photon_emax[-1])
        photon_egrid = np.asarray(photon_egrid, dtype=np.float64, order='C')

        if photon_egrid[0] <= 0.0:
            photon_egrid[0] = max(1e-30, 0.001 * photon_egrid[1])
            warnings.warn(
                f'non-positive photon energy bin detected in "{respfile}"; '
                'a small finite value will be used instead',
                Warning,
            )

        # check and read response matrix
        channel_number = response_header.get('DETCHANS', None)
        if channel_number is None:
            raise ValueError(
                f'keyword "DETCHANS" is not found in "{respfile}" header'
            )
        else:
            channel_number = int(channel_number)

        fchan_idx = response_data.names.index('F_CHAN') + 1
        # set the first channel number to 1 if not found
        first_chan = int(response_header.get(f'TLMIN{fchan_idx}', 1))

        channel = tuple(
            str(c) for c in range(first_chan, first_chan + channel_number)
        )

        n_grp = response_data['N_GRP']
        f_chan = response_data['F_CHAN'] - first_chan
        n_chan = response_data['N_CHAN']

        # if ndim == 1 and dtype is 'O', it is an array of array
        if f_chan.ndim == 1 and f_chan.dtype != np.dtype('O'):
            # f_chan is scalar in each row, make it an array
            f_chan = f_chan[:, None]

        if n_chan.ndim == 1 and n_chan.dtype != np.dtype('O'):
            # n_chan is scalar in each row, make it an array
            n_chan = n_chan[:, None]

        # the last channel numbers
        e_chan = f_chan + n_chan

        rows = np.hstack(
            tuple(np.full(round(n.sum()), i) for i, n in enumerate(n_chan))
        )
        cols = []
        for i in range(len(response_data)):
            n = int(n_grp[i])  # n channel subsets
            if n > 0:
                f = f_chan[i].astype(int)  # first channel numbers of subsets
                e = e_chan[i].astype(int)  # last channel numbers of subsets
                cols.extend(map(np.arange, f, e))
        cols = np.hstack(cols)

        matrix = response_data['MATRIX'].ravel()
        if matrix.dtype != np.dtype('O'):
            reduced_matrix = matrix
        else:
            reduced_matrix = np.hstack(matrix)

        # convert to native byteorder to be compatible with scipy.sparse
        reduced_matrix = to_native_byteorder(reduced_matrix)

        sparse_matrix = coo_array(
            (reduced_matrix, (rows, cols)),
            shape=(len(response_data), channel_number),
        )
        sparse_matrix.eliminate_zeros()

        return {
            'photon_egrid': photon_egrid,
            'channel_emin': channel_emin,
            'channel_emax': channel_emax,
            'channel': channel,
            'channel_type': channel_type,
            'response_matrix': sparse_matrix,
        }

    def _read_arf(self) -> NDArray | None:
        if self.ancrfile:
            with fits.open(self.ancrfile) as arf_hdul:
                return arf_hdul['SPECRESP'].data['SPECRESP']

    @staticmethod
    def _drop_zeros(
        photon_egrid: NDArray, sparse_matrix: coo_array
    ) -> tuple[NDArray, coo_array]:
        """Remove leading or trailing rows filled with 0 from the matrix."""
        nonzero_rows = np.unique(sparse_matrix.nonzero()[0])
        nonzero_mask = np.isin(range(sparse_matrix.shape[0]), nonzero_rows)
        zero_mask = np.bitwise_not(nonzero_mask)
        if zero_mask.any():
            n_entries = len(photon_egrid) - 1
            last_idx = len(photon_egrid) - 2
            idx = np.flatnonzero(zero_mask)
            diff = np.diff(idx)
            if len(diff) == 0:  # only one zero entry
                idx = idx[0]
                if idx in (0, last_idx):  # check if idx is leading or trailing
                    if idx == 0:
                        photon_egrid = photon_egrid[1:]
                    else:
                        photon_egrid = photon_egrid[:-1]
            else:
                splits = np.split(idx, np.nonzero(np.diff(idx) > 1)[0] + 1)
                zeros_mask2 = np.full(n_entries, False)
                for s in splits:
                    if np.isin(s, (0, last_idx)).any():
                        # only drop leading or trailing part of grids
                        zeros_mask2[s] = True

                elo = photon_egrid[:-1][~zeros_mask2]
                ehi = photon_egrid[1:][~zeros_mask2]
                photon_egrid = np.append(elo, ehi[-1])

        return photon_egrid, sparse_matrix.tocsr()[~zero_mask]

    @property
    def respfile(self) -> str:
        """The response file path."""
        return self._respfile

    @property
    def ancrfile(self) -> str:
        """The ancillary response file path."""
        return self._ancrfile


class PoissonFlagNotFoundError(RuntimeError):
    """Issued by ``POISSERR`` not found in spectrum header."""



"""Likelihood functions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, get_args

import jax
import jax.numpy as jnp
import numpyro
from jax import lax
from jax.experimental.sparse import BCSR
from jax.scipy.special import xlogy
from numpyro.distributions import Normal, Poisson
from numpyro.distributions.util import validate_sample

if TYPE_CHECKING:
    from collections.abc import Callable

    from elisa.data.base import FixedData
    from elisa.util.typing import (
        ArrayLike,
        JAXArray,
        ModelCompiledFn,
        ParamNameValMapping,
    )


# TODO:
#   It should be noted that 'lstat' does not have long run coverage property
#   for source estimation, which is probably due to the choice of conjugate
#   prior of Poisson background data.
#   'lstat' will be included here with a proper prior at some point.
Statistic = Literal['chi2', 'cstat', 'pstat', 'pgstat', 'wstat']

_STATISTIC_OPTIONS: frozenset[str] = frozenset(get_args(Statistic))
_STATISTIC_SPEC_NORMAL: frozenset[str] = frozenset({'chi2'})
_STATISTIC_BACK_NORMAL: frozenset[str] = frozenset({'pgstat'})
_STATISTIC_WITH_BACK: frozenset[str] = frozenset({'pgstat', 'wstat'})


def pgstat_background(
    s: ArrayLike,
    n: ArrayLike,
    b_est: ArrayLike,
    b_err: ArrayLike,
    a: ArrayLike,
) -> JAXArray:
    """Optimized background for PG-statistics given estimate of source counts.

    .. note::
        The optimized background here is always non-negative, which differs
        from XSPEC [1]_.

    Parameters
    ----------
    s : array_like
        Estimate of source counts.
    n : array_like
        Observed counts (source and background).
    b_est : array_like
        Estimate of background counts.
    b_err : array_like
        Uncertainty of background counts.
    a : float or array_like
        Exposure ratio between source and background observations.

    Returns
    -------
    JAXArray
        The profile background.

    References
    ----------
    .. [1] `XSPEC Manual Appendix B: Statistics in XSPEC <https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/XSappendixStatistics.html>`__.
    """
    variance = b_err * b_err
    e = jnp.array(b_est - a * variance)
    f = a * variance * n + e * s
    c = a * e - s
    d = jnp.sqrt(c * c + 4.0 * a * f)
    b = jnp.where(
        jnp.bitwise_or(jnp.greater_equal(e, 0.0), jnp.greater_equal(f, 0.0)),
        jnp.where(jnp.greater(n, 0.0), (c + d) / (2 * a), e),
        0.0,
    )
    return b


def wstat_background(
    s: ArrayLike,
    n_on: ArrayLike,
    n_off: ArrayLike,
    a: ArrayLike,
) -> JAXArray:
    """Optimized background for W-statistics [1]_ given the estimate of source.

    Parameters
    ----------
    s : array_like
        Estimate of source counts.
    n_on : array_like
        Observed source and background counts in "on" observation.
    n_off : array_like
        Observed background counts in "off" observation.
    a : array_like
        Exposure ratio between "on" and "off" observations.

    Returns
    -------
    JAXArray
        The profile background.

    References
    ----------
    .. [1] Wachter, K., Leach, R., & Kellogg, E. (1979). Parameter estimation
       in X-ray astronomy using maximum likelihood. ApJ, 230, 274–287.
    """
    c = a * (n_on + n_off) - (a + 1) * s
    d = jnp.sqrt(c * c + 4 * a * (a + 1) * n_off * s)
    b = jnp.where(
        jnp.equal(n_on, 0),
        n_off / (1 + a),
        jnp.where(
            jnp.equal(n_off, 0),
            jnp.where(
                jnp.less_equal(s, a / (a + 1) * n_on),
                n_on / (1 + a) - s / a,
                0.0,
            ),
            (c + d) / (2 * a * (a + 1)),
        ),
    )
    return b


class BetterNormal(Normal):
    @validate_sample
    def log_prob(self, value):
        value_scaled = (value - self.loc) / self.scale
        return -0.5 * value_scaled * value_scaled


class BetterPoisson(Poisson):
    @validate_sample
    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)

        if (
            self.is_sparse
            and not isinstance(value, jax.core.Tracer)
            and jnp.size(value) > 1
        ):
            broadcast = jnp.broadcast_to
            shape = lax.broadcast_shapes(self.batch_shape, jnp.shape(value))
            rate = broadcast(self.rate, shape).reshape(-1)
            nonzero = broadcast(jax.device_get(value) > 0, shape).reshape(-1)
            value = broadcast(value, shape).reshape(-1)
            sparse_value = value[nonzero]
            sparse_rate = rate[nonzero]
            tmp = xlogy(sparse_value, sparse_rate)
            gof = xlogy(sparse_value, sparse_value) - sparse_value
            return jnp.clip(
                jnp.asarray(-rate, dtype=jnp.result_type(float))
                .at[nonzero]
                .add(tmp - gof)
                .reshape(shape),
                max=0.0,
            )

        else:
            logp = xlogy(value, self.rate) - self.rate
            gof = xlogy(value, value) - value
            return jnp.clip(logp - gof, max=0.0)


def _get_resp_matrix(data: FixedData) -> JAXArray | BCSR:
    if data.response_sparse:
        return BCSR.from_scipy_sparse(data.sparse_matrix.T)
    else:
        return jnp.array(data.response_matrix.T, float)


def chi2(
    data: FixedData,
    model: ModelCompiledFn,
) -> Callable[[ParamNameValMapping, bool], None]:
    """S^2 statistic, Gaussian likelihood."""
    name = str(data.name)
    spec = jnp.array(data.net_counts, float)
    error = jnp.array(data.net_errors, float)
    photon_egrid = jnp.array(data.photon_egrid, float)
    channel_width = jnp.array(data.channel_width, float)
    resp_matrix = _get_resp_matrix(data)
    area_scale = jnp.array(data.area_scale, float)
    exposure = jnp.array(data.spec_exposure, float)

    def likelihood(
        params: ParamNameValMapping,
        predictive: bool = False,
    ) -> None:
        """Gaussian likelihood defined via numpyro primitives."""
        unfold = model(photon_egrid, params)
        unfold = jnp.clip(unfold, min=1e-300, max=1e300)
        source_rate = resp_matrix @ unfold * area_scale
        numpyro.deterministic(name, source_rate / channel_width)
        source_counts = source_rate * exposure
        source_counts = jnp.clip(source_counts, min=1e-30, max=1e15)
        spec_data = numpyro.primitives.mutable(f'{name}_Non_data', spec)
        spec_model = numpyro.deterministic(f'{name}_Non_model', source_counts)

        with numpyro.plate(f'{name}_plate', len(spec)):
            dist_on = BetterNormal(spec_model, error)
            numpyro.sample(
                name=f'{name}_Non',
                fn=dist_on,
                obs=None if predictive else spec_data,
            )

        # record log likelihood into chains to avoid re-computation
        if not predictive:
            loglike_on = numpyro.deterministic(
                name=f'{name}_Non_loglike', value=dist_on.log_prob(spec_data)
            )
            numpyro.deterministic(name=f'{name}_loglike', value=loglike_on)

    return likelihood


def cstat(
    data: FixedData,
    model: ModelCompiledFn,
) -> Callable[[ParamNameValMapping, bool], None]:
    """C-statistic, Poisson likelihood."""
    name = str(data.name)
    spec = jnp.array(data.spec_counts, float)
    photon_egrid = jnp.array(data.photon_egrid, float)
    channel_width = jnp.array(data.channel_width, float)
    resp_matrix = _get_resp_matrix(data)
    area_scale = jnp.array(data.area_scale, float)
    exposure = jnp.array(data.spec_exposure, float)

    def likelihood(
        params: ParamNameValMapping,
        predictive: bool = False,
    ) -> None:
        """Poisson likelihood defined via numpyro primitives."""
        unfold = model(photon_egrid, params)
        unfold = jnp.clip(unfold, min=1e-300, max=1e300)
        source_rate = resp_matrix @ unfold * area_scale
        numpyro.deterministic(name, source_rate / channel_width)
        source_counts = source_rate * exposure
        source_counts = jnp.clip(source_counts, min=1e-30, max=1e15)
        spec_data = numpyro.primitives.mutable(f'{name}_Non_data', spec)
        spec_model = numpyro.deterministic(f'{name}_Non_model', source_counts)

        with numpyro.plate(f'{name}_plate', len(spec)):
            dist_on = BetterPoisson(spec_model)
            numpyro.sample(
                name=f'{name}_Non',
                fn=dist_on,
                obs=None if predictive else spec_data,
            )

        # record log likelihood into chains to avoid re-computation
        if not predictive:
            loglike_on = numpyro.deterministic(
                name=f'{name}_Non_loglike', value=dist_on.log_prob(spec_data)
            )
            numpyro.deterministic(name=f'{name}_loglike', value=loglike_on)

    return likelihood


def pstat(
    data: FixedData,
    model: ModelCompiledFn,
) -> Callable[[ParamNameValMapping, bool], None]:
    """P-statistic, Poisson likelihood for data with a known background."""
    assert data.has_back, 'Data must have background'

    name = str(data.name)
    spec = jnp.array(data.spec_counts, float)
    back = jnp.array(data.back_counts, float)
    photon_egrid = jnp.array(data.photon_egrid, float)
    channel_width = jnp.array(data.channel_width, float)
    resp_matrix = _get_resp_matrix(data)
    area_scale = jnp.array(data.area_scale, float)
    exposure = jnp.array(data.spec_exposure, float)
    back_ratio = jnp.array(data.back_ratio, float)

    def likelihood(
        params: ParamNameValMapping,
        predictive: bool = False,
    ) -> None:
        """Poisson likelihood defined via numpyro primitives."""
        unfold = model(photon_egrid, params)
        unfold = jnp.clip(unfold, min=1e-300, max=1e300)
        source_rate = resp_matrix @ unfold * area_scale
        numpyro.deterministic(name, source_rate / channel_width)
        model_counts = source_rate * exposure + back_ratio * back
        model_counts = jnp.clip(model_counts, min=1e-30, max=1e15)
        spec_data = numpyro.primitives.mutable(f'{name}_Non_data', spec)
        spec_model = numpyro.deterministic(f'{name}_Non_model', model_counts)

        with numpyro.plate(f'{name}_plate', len(spec_data)):
            dist_on = BetterPoisson(spec_model)
            numpyro.sample(
                name=f'{name}_Non',
                fn=dist_on,
                obs=None if predictive else spec_data,
            )

        # record log likelihood into chains to avoid re-computation
        if not predictive:
            loglike_on = numpyro.deterministic(
                name=f'{name}_Non_loglike', value=dist_on.log_prob(spec_data)
            )
            numpyro.deterministic(name=f'{name}_loglike', value=loglike_on)

    return likelihood


def pgstat(
    data: FixedData,
    model: ModelCompiledFn,
) -> Callable[[ParamNameValMapping, bool], None]:
    """PG-statistic, Poisson likelihood for data and profile Gaussian
    likelihood for background.
    """
    assert data.has_back, 'Data must have background'

    name = str(data.name)
    spec = jnp.array(data.spec_counts, float)
    back = jnp.array(data.back_counts, float)
    back_error = jnp.array(data.back_errors, float)
    photon_egrid = jnp.array(data.photon_egrid, float)
    channel_width = jnp.array(data.channel_width, float)
    resp_matrix = _get_resp_matrix(data)
    area_scale = jnp.array(data.area_scale, float)
    exposure = jnp.array(data.spec_exposure, float)
    back_ratio = jnp.array(data.back_ratio, float)

    def likelihood(params: ParamNameValMapping, predictive: bool = False):
        """Poisson and Gaussian likelihood defined via numpyro primitives."""
        unfold = model(photon_egrid, params)
        unfold = jnp.clip(unfold, min=1e-300, max=1e300)
        source_rate = resp_matrix @ unfold * area_scale
        numpyro.deterministic(name, source_rate / channel_width)
        spec_data = numpyro.primitives.mutable(f'{name}_Non_data', spec)
        back_data = numpyro.primitives.mutable(f'{name}_Noff_data', back)
        source_counts = source_rate * exposure
        source_counts = jnp.clip(source_counts, min=1e-30, max=1e15)
        b = pgstat_background(
            source_counts, spec_data, back_data, back_error, back_ratio
        )
        spec_model = source_counts + back_ratio * b
        spec_model = numpyro.deterministic(f'{name}_Non_model', spec_model)
        back_model = numpyro.deterministic(f'{name}_Noff_model', b)

        with numpyro.plate(f'{name}_plate', len(spec_data)):
            dist_on = BetterPoisson(spec_model)
            dist_off = BetterNormal(back_model, back_error)
            numpyro.sample(
                name=f'{name}_Non',
                fn=dist_on,
                obs=None if predictive else spec_data,
            )
            numpyro.sample(
                name=f'{name}_Noff',
                fn=dist_off,
                obs=None if predictive else back_data,
            )

        # record log likelihood into chains to avoid re-computation
        if not predictive:
            loglike_on = numpyro.deterministic(
                name=f'{name}_Non_loglike', value=dist_on.log_prob(spec_data)
            )
            loglike_off = numpyro.deterministic(
                name=f'{name}_Noff_loglike', value=dist_off.log_prob(back_data)
            )
            numpyro.deterministic(
                name=f'{name}_loglike', value=loglike_on + loglike_off
            )

    return likelihood


def wstat(
    data: FixedData,
    model: ModelCompiledFn,
) -> Callable[[ParamNameValMapping, bool], None]:
    """W-statistic, i.e. Poisson likelihood for data and profile Poisson
    likelihood for background.
    """
    assert data.has_back, 'Data must have background'

    name = str(data.name)
    spec = jnp.array(data.spec_counts, float)
    back = jnp.array(data.back_counts, float)
    photon_egrid = jnp.array(data.photon_egrid, float)
    channel_width = jnp.array(data.channel_width, float)
    resp_matrix = _get_resp_matrix(data)
    area_scale = jnp.array(data.area_scale, float)
    exposure = jnp.array(data.spec_exposure, float)
    back_ratio = jnp.array(data.back_ratio, float)

    def likelihood(params: ParamNameValMapping, predictive: bool = False):
        """Poisson and Poisson likelihood defined via numpyro primitives."""
        unfold = model(photon_egrid, params)
        unfold = jnp.clip(unfold, min=1e-300, max=1e300)
        source_rate = resp_matrix @ unfold * area_scale
        numpyro.deterministic(name, source_rate / channel_width)
        spec_data = numpyro.primitives.mutable(f'{name}_Non_data', spec)
        back_data = numpyro.primitives.mutable(f'{name}_Noff_data', back)
        source_counts = source_rate * exposure
        source_counts = jnp.clip(source_counts, min=1e-30, max=1e15)
        b = wstat_background(source_counts, spec_data, back_data, back_ratio)
        model_counts = source_counts + back_ratio * b
        spec_model = numpyro.deterministic(f'{name}_Non_model', model_counts)
        back_model = numpyro.deterministic(f'{name}_Noff_model', b)

        with numpyro.plate(f'{name}_plate', len(spec_data)):
            dist_on = BetterPoisson(spec_model)
            dist_off = BetterPoisson(back_model)
            numpyro.sample(
                name=f'{name}_Non',
                fn=dist_on,
                obs=None if predictive else spec_data,
            )
            numpyro.sample(
                name=f'{name}_Noff',
                fn=dist_off,
                obs=None if predictive else back_data,
            )

        # record log likelihood into chains to avoid re-computation
        if not predictive:
            loglike_on = numpyro.deterministic(
                name=f'{name}_Non_loglike', value=dist_on.log_prob(spec_data)
            )
            loglike_off = numpyro.deterministic(
                name=f'{name}_Noff_loglike', value=dist_off.log_prob(back_data)
            )
            numpyro.deterministic(
                name=f'{name}_loglike', value=loglike_on + loglike_off
            )

    return likelihood



"""Methods for grouping spectrum."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy.special import xlogy

if TYPE_CHECKING:
    NDArray = np.ndarray


def group_const(n: int, c: int) -> tuple[NDArray, bool]:
    """Group data by containing `c` channels in each group.

    Parameters
    ----------
    n : int
        Number of channels.
    c : int
        Number of channels in each group.

    Returns
    -------
    flag : ndarray
        Grouping flag.
    success: bool
        Whether the scale is met for all grouped channels.
    """
    n = int(n)
    c = int(c)
    assert c > 0
    flag = np.full(n, -1, dtype=int)
    flag[::c] = 1
    r = n % c
    if r:  # Ensure the last group has at least `c` channels.
        flag[-r] = -1
    return flag, True


def group_min(data: NDArray, n: int) -> tuple[NDArray, bool]:
    """Group data by containing at least `n` counts in each channel.

    Parameters
    ----------
    data: array_like
        Counts array.
    n: int
        Minimum number of counts in each channel after grouping.

    Returns
    -------
    flag : ndarray
        Grouping flag.
    success: bool
        Whether the scale is met for all grouped channels.
    """
    n = int(n)
    assert n > 0
    nd = len(data)
    idx = np.empty(nd, np.int64)
    idx[0] = 0

    group_counts = 0
    ng = 1
    imax = nd - 1

    for i, di in enumerate(data):
        group_counts += di

        if i == imax:
            if group_counts < n and ng > 1:
                # if the last group does not have enough counts,
                # then combine the last two groups to ensure all
                # groups meet the count requirement
                ng -= 1

            break

        if group_counts >= n:
            idx[ng] = i + 1
            ng += 1
            group_counts = 0

    idx = idx[:ng]
    if np.all(np.add.reduceat(data, idx) >= n):
        success = True
    else:
        success = False

    flag = np.full(nd, -1, dtype=int)
    flag[idx] = 1

    return flag, success


# TODO: the total number of resolution elements R should be given as argument
def _calc_optimal_binning(fwhm: NDArray, counts: NDArray) -> NDArray:
    """Calculate the optimal binning for each channel.

    .. note::
        This is translated from the ``heasp``.
    """
    # The optimal binning from Kaastra & Bleeker 2016, A&A 587, A151.
    # The binning delta/FWHM is defined as
    #    1  if x <= 2.119 otherwise
    #    (0.08 + 7.0/x + 1.8/x^2)/(1 + 5.9/x)
    # where
    #    x = ln[N_r(1 + 0.2 ln R)]
    # and N_r is the number of counts per resolution element and R is the total
    # number of resolution elements.
    Nchan = len(fwhm)

    # Estimate the total number of resolution elements. Since this enters
    # as ln(0.2ln R) it can be a rough estimate. See eqn B.1 of K&B.
    logR = np.log(1 + np.sum(1.0 / fwhm))

    # Calculate the number of counts within the FWHM for each channel then
    # multiply by 1.314 to get the number of counts per resolution element.
    # This assumes the response is gaussian - I could improve this by also
    # including a vector with the fraction within the FWHM for each channel
    # but this is probably not going to make a significant difference
    Nr = np.zeros(Nchan)
    for i in range(Nchan):
        fwhm_i = fwhm[i]
        low = max(0, int(np.round(i - 0.5 * fwhm_i)))
        high = min(Nchan, int(np.round(i + 0.5 * fwhm_i)))
        Nr[i] = 1.314 * counts[low : high + 1].sum()

    # Calculate the optimal bin size at each channel
    b = np.array(fwhm, dtype=np.float64)
    mask = Nr > np.exp(2.119) / (1 + 0.2 * logR)
    x = np.log(Nr[mask] * (1.0 + 0.2 * logR))
    b[mask] *= (0.08 * x + 7.0 + 1.8 / x) / (x + 5.9)
    bint = b.astype(np.int64)
    bint[bint < 1] = 1
    return bint


def group_opt(
    fwhm: NDArray,
    net_counts: NDArray,
    bin_counts: NDArray | None = None,
    n: int | None = None,
) -> tuple[NDArray, bool]:
    """Optimal binning of the spectrum data [1]_.

    Parameters
    ----------
    fwhm : ndarray
        FWHM of the channel.
    net_counts : ndarray
        Net counts of the channel.
    bin_counts : ndarray, optional
        Additional counts to be grouped.
    n : int, optional
        Grouping scale of the `bin_counts`.

    Returns
    -------
    flag : ndarray
        Grouping flag.
    success: bool
        Whether the scale is met for all grouped channels.

    References
    ----------
    .. [1] Kaastra & Bleeker 2016, A&A 587, A151
    """
    assert len(fwhm) == len(net_counts)
    nchan = len(fwhm)

    # Calculate the optimal binsize for each channel based on
    # the FWHM and the counts.
    bint = _calc_optimal_binning(fwhm, net_counts)

    # Initialize the grouping flag
    if n is not None:
        n = int(n)
    else:
        n = 0
    if bin_counts is not None:
        bin_counts = np.array(bin_counts)
    min_flag = (n > 0) and (bin_counts is not None)
    idx = np.empty(nchan, dtype=np.int64)
    ng = 0

    i = 0
    imax = nchan - 1
    while i <= imax:
        idx[ng] = i
        ng += 1

        j = min(imax, i + bint[i] - 1)
        k = np.arange(i + 1, j + 1)
        if k.size:
            j = np.minimum(j, np.min(k + bint[k] - 1))

        # Ensure minimum number of counts and extend the bin
        # if necessary to abide that constraint
        if min_flag:
            cts = bin_counts[i : j + 1].sum()
            if cts < n:
                mask = bin_counts[j + 1 :].cumsum() >= n - cts
                if np.any(mask):
                    # the smallest j for bin_counts[i:j + 1].sum() >= n
                    j += np.flatnonzero(mask)[0] + 1
                else:
                    # if the last group does not have enough counts,
                    # then combine the last two groups to ensure all
                    # groups meet the count requirement
                    if ng > 1:
                        ng -= 1
                    break

        if j > imax:
            # If the left channels are not enough, then group them with
            # the previous channels group.
            if (j - imax) / (j - i) > 1 / 3 and ng > 1:
                ng -= 1
            break

        i = j + 1

    idx = idx[:ng]
    flag = np.full(nchan, -1, dtype=int)
    flag[idx] = 1

    if min_flag and np.any(np.add.reduceat(bin_counts, idx) < n):
        success = False
    else:
        success = True

    return flag, success


def significance_lima(
    n_on: float | NDArray,
    n_off: float | NDArray,
    a: float | NDArray,
) -> NDArray:
    """Significance using the formula of Li & Ma 1983."""
    n_on = np.asarray(n_on)
    n_off = np.asarray(n_off)

    term1 = np.zeros_like(n_on).astype(float)
    mask = n_on > 0.0
    if mask.any():
        term1[mask] = xlogy(
            n_on[mask], (1.0 + a) / a * n_on[mask] / (n_on[mask] + n_off[mask])
        )

    term2 = np.zeros_like(n_on).astype(float)
    mask = n_off > 0.0
    if mask.any():
        term2[mask] = xlogy(
            n_off[mask], (1.0 + a) * n_off[mask] / (n_on[mask] + n_off[mask])
        )
    sign = np.where(n_on >= a * n_off, 1.0, -1.0)
    return sign * np.sqrt(2.0 * (term1 + term2))


def significance_gv(
    n: float | NDArray,
    b: float | NDArray,
    s: float | NDArray,
    a: float | NDArray,
) -> NDArray:
    """Significance using the formula of Vianello 2018."""
    n = np.asarray(n)
    b = np.asarray(b)
    s = np.asarray(s)

    b = b * a
    s = s * a
    s2 = s * s
    s4 = s2 * s2
    b0_mle = 0.5 * (b - s2 + np.sqrt(b * b - 2 * b * s2 + 4 * n * s2 + s4))
    b0_mle = np.clip(b0_mle, 0, None)

    term1 = np.zeros_like(n).astype(float)
    mask = n > 0.0
    if mask.any():
        n_mask = n[mask]
        b0_mle_mask = b0_mle[mask]
        term1[mask] = xlogy(n_mask, n_mask / b0_mle_mask)
    term1 += b0_mle - n

    term2 = np.square(b - b0_mle) / (2 * s2)
    sign = np.where(n >= b, 1.0, -1.0)
    return sign * np.sqrt(2.0 * (term1 + term2))


def group_optsig_normal(
    fwhm: NDArray,
    net_counts: NDArray,
    counts: NDArray,
    errors: NDArray,
    sig: int | float,
) -> tuple[NDArray, bool]:
    """Optimal binning with an extra requirement of a minimum significance.

    Parameters
    ----------
    fwhm : ndarray
        FWHM of the channel.
    net_counts : ndarray
        Net counts of the channel.
    counts : ndarray
        Counts of the channel.
    errors : ndarray
        Uncertainty of the counts.
    sig : float
        Significance threshold.

    Returns
    -------
    flag : ndarray
        Grouping flag.
    success: bool
        Whether the scale is met for all grouped channels.
    """
    assert len(fwhm) == len(net_counts) == len(counts) == len(errors)
    sig = float(sig)
    assert sig > 0.0
    nchan = len(fwhm)

    # Calculate the optimal binsize for each channel based on
    # the FWHM and the counts.
    bint = _calc_optimal_binning(fwhm, net_counts)

    # Initialize the grouping flag
    idx = np.empty(nchan, dtype=np.int64)
    ng = 0

    i = 0
    imax = nchan - 1
    while i <= imax:
        idx[ng] = i
        ng += 1

        j = min(imax, i + bint[i] - 1)
        k = np.arange(i + 1, j + 1)
        if k.size:
            j = np.minimum(j, np.min(k + bint[k] - 1))

        # Ensure minimum significance and extend the bin if necessary to abide
        # that constraint
        cts = counts[i : j + 1].sum()
        err = np.sqrt(np.sum(np.square(errors[i : j + 1])))
        if cts - sig * err < 0.0 or err == 0.0:
            cts = cts + counts[j + 1 :].cumsum()
            err = err + np.sqrt(np.cumsum(np.square(errors[j + 1 :])))
            mask = (cts - sig * err >= 0.0) & (err > 0.0)
            if np.any(mask):
                # the smallest j for the significance threshold
                j += np.flatnonzero(mask)[0] + 1
            else:
                # if the last group does not have a nominal significance,
                # then combine the last two groups to ensure all
                # groups meet the count requirement
                if ng > 1:
                    ng -= 1
                break

        if j > imax:
            # If the left channels are not enough, then group them with
            # the previous channels group.
            if (j - imax) / (j - i) > 1 / 3 and ng > 1:
                ng -= 1
            break

        i = j + 1

    idx = idx[:ng]
    flag = np.full(nchan, -1, dtype=int)
    flag[idx] = 1

    cts = np.add.reduceat(counts, idx)
    err = np.sqrt(np.add.reduceat(errors * errors, idx))
    if np.all(cts - sig * err >= 0.0):
        success = True
    else:
        success = False

    return flag, success


def group_optsig_lima(
    fwhm: NDArray,
    net_counts: NDArray,
    n_on: NDArray,
    n_off: NDArray,
    a: float,
    sig: int | float,
) -> tuple[NDArray, bool]:
    """Optimal binning with an extra requirement of a minimum significance.

    .. note::
        The formula of Li & Ma 1983 is used to calculate the significance.

    Parameters
    ----------
    fwhm : ndarray
        FWHM of the channel.
    net_counts : ndarray
        Net counts of the channel.
    n_on : ndarray
        Counts array of on measurement.
    n_off : ndarray
        Counts array of off measurement.
    a : float
        Ratio of the on and off exposure.
    sig : int or float
        Significance threshold.

    Returns
    -------
    flag : ndarray
        Grouping flag.
    success: bool
        Whether the scale is met for all grouped channels.
    """
    assert len(fwhm) == len(net_counts) == len(n_on) == len(n_off)
    assert a > 0.0
    sig = float(sig)
    assert sig > 0.0
    nchan = len(fwhm)

    # Calculate the optimal binsize for each channel based on
    # the FWHM and the counts.
    bint = _calc_optimal_binning(fwhm, net_counts)

    # Initialize the grouping flag
    idx = np.empty(nchan, dtype=np.int64)
    ng = 0

    i = 0
    imax = nchan - 1
    while i <= imax:
        idx[ng] = i
        ng += 1

        j = min(imax, i + bint[i] - 1)
        k = np.arange(i + 1, j + 1)
        if k.size:
            j = np.minimum(j, np.min(k + bint[k] - 1))

        # Ensure minimum significance and extend the bin if necessary to abide
        # that constraint
        on = n_on[i : j + 1].sum()
        off = n_off[i : j + 1].sum()
        if significance_lima(on, off, a) < sig:
            on = on + n_on[j + 1 :].cumsum()
            off = off + n_off[j + 1 :].cumsum()
            mask = significance_lima(on, off, a) >= sig
            if np.any(mask):
                # the smallest j for the significance threshold
                j += np.flatnonzero(mask)[0] + 1
            else:
                # if the last group does not have a nominal significance,
                # then combine the last two groups to ensure all
                # groups meet the count requirement
                if ng > 1:
                    ng -= 1

        if j > imax:
            # If the left channels are not enough, then group them with
            # the previous channels group.
            if (j - imax) / (j - i) > 1 / 3 and ng > 1:
                ng -= 1
            break

        i = j + 1

    idx = idx[:ng]
    flag = np.full(nchan, -1, dtype=int)
    flag[idx] = 1

    on = np.add.reduceat(n_on, idx)
    off = np.add.reduceat(n_off, idx)
    if np.all(significance_lima(on, off, a) >= sig):
        success = True
    else:
        success = False

    return flag, success


def group_optsig_gv(
    fwhm: NDArray,
    net_counts: NDArray,
    n: NDArray,
    b: NDArray,
    s: NDArray,
    a: float,
    sig: int | float,
) -> tuple[NDArray, bool]:
    """Optimal binning with an extra requirement of a minimum significance.

    .. note::
        The formula of Vianello 2018 is used to calculate the significance.

    Parameters
    ----------
    fwhm : ndarray
        FWHM of the channel.
    net_counts : ndarray
        Net counts of the channel.
    n : ndarray
        Counts array of on measurement.
    b : ndarray
        Estimate of background counts.
    s : ndarray
        Uncertainty of the background estimate.
    a : float
        Ratio of the on and off exposure.
    sig : float
        Significance threshold.

    Returns
    -------
    flag : ndarray
        Grouping flag.
    success: bool
        Whether the scale is met for all grouped channels.
    """
    assert len(fwhm) == len(net_counts) == len(n) == len(b) == len(s)
    assert a > 0.0
    sig = float(sig)
    assert sig > 0.0
    nchan = len(fwhm)

    # Calculate the optimal binsize for each channel based on
    # the FWHM and the counts.
    bint = _calc_optimal_binning(fwhm, net_counts)

    # Initialize the grouping flag
    idx = np.empty(nchan, dtype=np.int64)
    ng = 0

    i = 0
    imax = nchan - 1
    while i <= imax:
        idx[ng] = i
        ng += 1

        j = min(imax, i + bint[i] - 1)
        k = np.arange(i + 1, j + 1)
        if k.size:
            j = np.minimum(j, np.min(k + bint[k] - 1))

        # Ensure minimum significance and extend the bin if necessary to abide
        # that constraint
        n_ = n[i : j + 1].sum()
        b_ = b[i : j + 1].sum()
        s_ = np.sqrt(np.sum(np.square(s[i : j + 1])))
        if significance_gv(n_, b_, s_, a) < sig:
            n_ = n_ + n[j + 1 :].cumsum()
            b_ = b_ + b[j + 1 :].cumsum()
            s_ = s_ + np.sqrt(np.cumsum(np.square(s[j + 1 :])))
            mask = significance_gv(n_, b_, s_, a) >= sig
            if np.any(mask):
                # the smallest j for the significance threshold
                j += np.flatnonzero(mask)[0] + 1
            else:
                # if the last group does not have a nominal significance,
                # then combine the last two groups to ensure all
                # groups meet the count requirement
                if ng > 1:
                    ng -= 1

        if j > imax:
            # If the left channels are not enough, then group them with
            # the previous channels group.
            if (j - imax) / (j - i) > 1 / 3 and ng > 1:
                ng -= 1
            break

        i = j + 1

    idx = idx[:ng]
    flag = np.full(nchan, -1, dtype=int)
    flag[idx] = 1

    n_ = np.add.reduceat(n, idx)
    b_ = np.add.reduceat(b, idx)
    s_ = np.sqrt(np.add.reduceat(s * s, idx))
    if np.all(significance_gv(n_, b_, s_, a) >= sig):
        success = True
    else:
        success = False

    return flag, success


def group_sig_normal(
    count: NDArray,
    error: NDArray,
    sig: int | float,
) -> tuple[NDArray, bool]:
    """Group data by limiting the significance is greater than `sig`.

    Parameters
    ----------
    count : ndarray
        Counts array.
    error : ndarray
        Uncertainty of counts.
    sig : int or float
        Significance threshold.

    Returns
    -------
    flag : ndarray
        Grouping flag.
    success: bool
        Whether the scale is met for all grouped channels.
    """
    assert len(count) == len(error)
    sig = float(sig)
    assert sig > 0.0

    nd = len(count)
    idx = np.empty(nd, np.int64)
    idx[0] = 0
    ng = 1
    imax = nd - 1

    group_count = 0.0
    group_variance = 0.0
    for i, (d, e) in enumerate(zip(count, error, strict=True)):
        group_count += d
        group_variance += e * e
        x = group_count - sig * np.sqrt(group_variance)

        if i == imax:
            if (x < 0.0 or group_variance == 0.0) and ng > 1:
                # if the last group does not have a nominal significance,
                # then combine the last two groups to ensure all
                # groups meet the count requirement
                ng -= 1

            break

        if x >= 0.0 and group_variance > 0.0:
            idx[ng] = i + 1
            ng += 1
            group_count = 0.0
            group_variance = 0.0

    idx = idx[:ng]
    group_count = np.add.reduceat(count, idx)
    group_error = np.sqrt(np.add.reduceat(error * error, idx))
    if np.all(group_count - sig * group_error >= 0):
        success = True
    else:
        success = False

    flag = np.full(nd, -1, dtype=int)
    flag[idx] = 1

    return flag, success


def group_sig_lima(
    n_on: NDArray,
    n_off: NDArray,
    a: float,
    sig: float,
) -> tuple[NDArray, bool]:
    """Group data by limiting the significance is greater than `sig`.

    .. note::
        The formula of Li & Ma 1983 is used to calculate the significance.

    Parameters
    ----------
    n_on : ndarray
        Counts array of on measurement.
    n_off : ndarray
        Counts array of off measurement.
    a : float
        Ratio of the on and off exposure.
    sig : int or float
        Significance threshold.

    Returns
    -------
    flag : ndarray
        Grouping flag.
    success: bool
        Whether the scale is met for all grouped channels.
    """
    assert len(n_on) == len(n_off)
    assert a > 0.0
    sig = float(sig)
    assert sig > 0.0

    nd = len(n_on)
    idx = np.empty(nd, np.int64)
    idx[0] = 0
    ng = 1
    imax = nd - 1

    group_on = 0.0
    group_off = 0.0
    for i, (j, k) in enumerate(zip(n_on, n_off, strict=True)):
        group_on += j
        group_off += k
        group_sig = significance_lima(group_on, group_off, a)

        if i == imax:
            if group_sig < sig and ng > 1:
                # if the last group does not have a nominal significance,
                # then combine the last two groups to ensure all
                # groups meet the count requirement
                ng -= 1

            break

        if group_sig >= sig:
            idx[ng] = i + 1
            ng += 1
            group_on = 0.0
            group_off = 0.0

    idx = idx[:ng]
    group_on = np.add.reduceat(n_on, idx)
    group_off = np.add.reduceat(n_off, idx)
    if np.all(significance_lima(group_on, group_off, a) >= sig):
        success = True
    else:
        success = False

    flag = np.full(nd, -1, dtype=int)
    flag[idx] = 1

    return flag, success


def group_sig_gv(
    n: NDArray,
    b: NDArray,
    s: NDArray,
    a: float,
    sig: float,
) -> tuple[NDArray, bool]:
    """Group data by limiting the significance is greater than `sig`.

    .. note::
        The formula of Vianello 2018 is used to calculate the significance.

    Parameters
    ----------
    n : ndarray
        Counts array of on measurement.
    b : ndarray
        Estimate of background counts.
    s : ndarray
        Uncertainty of the background estimate.
    a : float
        Ratio of the on and off exposure.
    sig : float
        Significance threshold.

    Returns
    -------
    flag : ndarray
        Grouping flag.
    success: bool
        Whether the scale is met for all grouped channels.
    """
    assert len(n) == len(b) == len(s)
    assert a > 0.0
    sig = float(sig)
    assert sig > 0.0

    nd = len(n)
    idx = np.empty(nd, np.int64)
    idx[0] = 0
    ng = 1
    imax = nd - 1

    group_n = 0.0
    group_b = 0.0
    group_var = 0.0
    for i, (ni, bi, si) in enumerate(zip(n, b, s, strict=True)):
        group_n += ni
        group_b += bi
        group_var += si * si
        group_sig = significance_gv(group_n, group_b, np.sqrt(group_var), a)

        if i == imax:
            if group_sig < sig and ng > 1:
                # if the last group does not have a nominal significance,
                # then combine the last two groups to ensure all
                # groups meet the count requirement
                ng -= 1

            break

        if group_sig >= sig:
            idx[ng] = i + 1
            ng += 1
            group_n = 0.0
            group_b = 0.0
            group_var = 0.0

    idx = idx[:ng]
    group_n = np.add.reduceat(n, idx)
    group_b = np.add.reduceat(b, idx)
    group_var = np.sqrt(np.add.reduceat(s * s, idx))
    group_sig = significance_gv(group_n, group_b, np.sqrt(group_var), a)
    if np.all(group_sig >= sig):
        success = True
    else:
        success = False

    flag = np.full(nd, -1, dtype=int)
    flag[idx] = 1

    return flag, success



"""Data classes for plotting."""

from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cache, wraps
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
import numpy as np
import scipy.stats as stats
from jax.experimental.mesh_utils import create_device_mesh
from jax.experimental.shard_map import shard_map
from jax.sharding import Mesh, PartitionSpec

from elisa.infer.likelihood import (
    _STATISTIC_BACK_NORMAL,
    _STATISTIC_SPEC_NORMAL,
    _STATISTIC_WITH_BACK,
)
from elisa.plot.residuals import (
    pearson_residuals,
    pit_poisson,
    pit_poisson_normal,
    pit_poisson_poisson,
    quantile_residuals_poisson,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any, Literal

    from xarray import DataArray

    from elisa.infer.results import (
        BootstrapResult,
        FitResult,
        MLEResult,
        PosteriorResult,
        PPCResult,
    )
    from elisa.util.typing import Array, NumPyArray


def _cache_method(bound_method: Callable) -> Callable:
    """Cache instance method."""
    return cache(bound_method)


def _cache_method_with_check(
    instance: Any, bound_method: Callable, check_fields: Sequence[str]
) -> Callable:
    """Cache instance method with computation dependency check."""

    def get_id():
        return {field: id(getattr(instance, field)) for field in check_fields}

    cached_method = cache(bound_method)
    old_id = get_id()

    @wraps(bound_method)
    def _(*args, **kwargs):
        if (new_id := get_id()) != old_id:
            cached_method.cache_clear()
            old_id.update(new_id)
        return cached_method(*args, **kwargs)

    return _


def _get_cached_method_decorator(storage: list):
    def decorator(method: Callable):
        storage.append(method.__name__)
        return method

    return decorator


def _get_cached_method_with_check_decorator(
    storage: list, check_fields: str | Sequence[str]
):
    if isinstance(check_fields, str):
        check_fields = [check_fields]
    else:
        check_fields = list(check_fields)

    def decorator(method: Callable):
        name = method.__name__
        storage.append((name, check_fields))
        return method

    return decorator


class PlotData(ABC):
    """Base class for data used in plotting."""

    _cached_method: list[str]
    _cached_method_with_check: list[tuple[str, list[str]]]
    _unfolded_model_fn: dict[str, Callable]
    _ph_egrid: NumPyArray | None = None

    def __init__(self, name: str, result: FitResult, seed: int):
        self.name = str(name)
        self.result = result
        self.seed = seed
        self.data = result._helper.data[self.name]
        self.statistic = result._helper.statistic[self.name]

        for f in self._cached_method:
            method = getattr(self, f)
            setattr(self, f, _cache_method(method))

        for f, fields in self._cached_method_with_check:
            method = getattr(self, f)
            setattr(self, f, _cache_method_with_check(self, method, fields))

        model = self.result._helper.model[self.name]
        self._unfolded_model_fn = {
            'ne': jax.jit(lambda e, p: model.ne(e, p, comps=False)),
            'ene': jax.jit(lambda e, p: model.ene(e, p, comps=False)),
            'eene': jax.jit(lambda e, p: model.eene(e, p, comps=False)),
            'ne_comps': jax.jit(lambda e, p: model.ne(e, p, comps=True)),
            'ene_comps': jax.jit(lambda e, p: model.ene(e, p, comps=True)),
            'eene_comps': jax.jit(lambda e, p: model.eene(e, p, comps=True)),
        }

    @property
    def channel(self) -> NumPyArray:
        return self.data.channel

    @property
    def channel_emin(self) -> NumPyArray:
        return self.data.channel_emin

    @property
    def channel_emax(self) -> NumPyArray:
        return self.data.channel_emax

    @property
    def channel_emid(self) -> NumPyArray:
        return self.data.channel_emid

    @property
    def channel_width(self) -> NumPyArray:
        return self.data.channel_width

    @property
    def channel_emean(self) -> NumPyArray:
        return self.data.channel_emean

    @property
    def channel_errors(self) -> NumPyArray:
        return self.data.channel_errors

    @property
    def photon_egrid(self) -> NumPyArray:
        if self._ph_egrid is not None:
            return self._ph_egrid

        ph_egrid = self.data.photon_egrid
        mask = np.bitwise_and(
            self.channel_emin[0] <= ph_egrid, ph_egrid <= self.channel_emax[-1]
        )
        self._ph_egrid = ph_egrid[mask]
        return self._ph_egrid

    @property
    def spec_counts(self) -> Array:
        return self.data.spec_counts

    @property
    def spec_errors(self) -> Array:
        return self.data.spec_errors

    @property
    def back_ratio(self) -> float | Array:
        return self.data.back_ratio

    @property
    def back_counts(self) -> Array | None:
        return self.data.back_counts

    @property
    def back_errors(self) -> Array | None:
        return self.data.back_errors

    @property
    def net_counts(self) -> Array:
        return self.data.net_counts

    @property
    def net_errors(self) -> Array:
        return self.data.net_errors

    @property
    def ndata(self) -> int:
        return len(self.data.channel)

    @property
    def ce_data(self) -> Array:
        return self.data.ce

    @property
    def ce_errors(self) -> Array:
        return self.data.ce_errors

    @property
    @abstractmethod
    def ce_model(self) -> Array:
        """Point estimate of the folded source model."""
        pass

    @abstractmethod
    def ce_model_ci(self, cl: float = 0.683) -> Array | None:
        """Confidence/Credible intervals of the folded source model."""
        pass

    @property
    def has_comps(self) -> bool:
        return self.result._helper.model[self.name].has_comps

    @property
    def params_dist(self) -> dict[str, Array] | None:
        return self.result._params_dist

    def _unfolded_model(
        self,
        mtype: Literal['ne', 'ene', 'eene'],
        egrid: Array,
        params: dict,
        comps: bool,
    ) -> Array | dict:
        assert mtype in {'ne', 'ene', 'eene'}
        fn = self._unfolded_model_fn[f'{mtype}_comps' if comps else mtype]
        if len(np.shape(list(params.values())[0])) != 0:
            devices = create_device_mesh((jax.local_device_count(),))
            mesh = Mesh(devices, axis_names=('i',))
            p = PartitionSpec()
            pi = PartitionSpec('i')
            fn = shard_map(
                f=fn,
                mesh=mesh,
                in_specs=(p, pi),
                out_specs=pi,
                check_rep=False,
            )
        return jax.device_get(fn(egrid, params))

    @abstractmethod
    def unfolded_model(
        self,
        mtype: Literal['ne', 'ene', 'eene'],
        egrid: Array,
        params: dict,
        comps: bool,
        cl: float | Array | None = None,
    ) -> Array | dict:
        pass

    @abstractmethod
    def pit(self) -> tuple:
        """Probability integral transform."""
        pass

    @abstractmethod
    def residuals(
        self,
        rtype: Literal['rd', 'rp', 'rq'],
        seed: int | None,
        random_quantile: bool,
        mle: bool,
    ) -> Array | tuple[Array, bool | Array, bool | Array]:
        """Residuals between the data and the fitted models."""
        pass

    @abstractmethod
    def residuals_sim(
        self,
        rtype: Literal['rd', 'rp', 'rq'],
        seed: int | None,
        random_quantile: bool,
    ) -> Array | None:
        """Residuals bootstrap/ppc samples."""
        pass

    @abstractmethod
    def residuals_ci(
        self,
        rtype: Literal['rd', 'rp', 'rq'],
        cl: float,
        seed: int | None,
        random_quantile: bool,
        with_sign: bool,
    ) -> Array | None:
        """Confidence/Credible intervals of the residuals."""
        pass


_cached_method = []
_cached_method_with_check = []
_to_cached_method = _get_cached_method_decorator(_cached_method)
_to_cached_method_with_check = _get_cached_method_with_check_decorator(
    _cached_method_with_check, 'boot'
)


class MLEPlotData(PlotData):
    result: MLEResult
    _cached_method = _cached_method
    _cached_method_with_check = _cached_method_with_check

    @property
    def boot(self) -> BootstrapResult:
        return self.result._boot

    @property
    def params_mle(self) -> dict[str, Array]:
        return {k: v[0] for k, v in self.result._mle.items()}

    def get_model_mle(self, name: str) -> Array:
        return self.result._model_values[name]

    def get_model_boot(self, name: str) -> Array | None:
        boot = self.boot
        if boot is None:
            return None
        else:
            return boot.models[name]

    def get_data_boot(self, name: str) -> Array | None:
        boot = self.boot
        if boot is None:
            return None
        else:
            return boot.data[name]

    @property
    def ce_model(self) -> Array:
        return self.get_model_mle(self.name)

    @_to_cached_method_with_check
    def ce_model_ci(self, cl: float = 0.683) -> Array | None:
        if self.boot is None:
            return None

        assert 0.0 < cl < 1.0
        ci = np.quantile(
            self.get_model_boot(self.name),
            q=0.5 + cl * np.array([-0.5, 0.5]),
            axis=0,
        )
        return ci

    def unfolded_model(
        self,
        mtype: Literal['ne', 'ene', 'eene'],
        egrid: Array | None,
        params: dict | None,
        comps: bool,
        cl: float | Array | None = None,
    ) -> tuple[Array | dict, Array | dict | None]:
        assert mtype in {'ne', 'ene', 'eene'}
        if cl is not None:
            cl = np.atleast_1d(cl).astype(float)
            assert np.all(0.0 < cl) and np.all(cl < 1.0)
        params = {} if params is None else dict(params)
        comps = comps and self.has_comps

        egrid = jnp.asarray(egrid, float)

        params_mle = self.params_mle | params
        model_mle = self._unfolded_model(mtype, egrid, params_mle, comps)

        params_boot = self.params_dist
        if cl is None or params_boot is None:
            return model_mle, None
        else:
            n = [i.size for i in params_boot.values()][0]
            if params:
                params = {k: jnp.full(n, v) for k, v in params.items()}
                params_boot = params_boot | params
            model_boot = self._unfolded_model(mtype, egrid, params_boot, comps)
            q = 0.5 + cl[:, None] * np.array([-0.5, 0.5])
            if comps:
                ci = {
                    k: np.quantile(v, q, axis=0) for k, v in model_boot.items()
                }
            else:
                ci = np.quantile(model_boot, q, axis=0)
            return model_mle, ci

    @property
    def sign(self) -> dict[str, Array | None]:
        """Sign of the difference between the data and the fitted models."""
        return {'mle': self._sign_mle(), 'boot': self._sign_boot()}

    @_to_cached_method
    def _sign_mle(self) -> Array:
        return np.where(self.ce_data >= self.ce_model, 1.0, -1.0)

    @_to_cached_method_with_check
    def _sign_boot(self) -> Array | None:
        boot = self.get_model_boot(self.name)
        if boot is not None:
            boot = np.where(self.get_data_boot(self.name) >= boot, 1.0, -1.0)
        return boot

    def model(
        self,
        on_off: Literal['on', 'off'],
        mtype: Literal['mle', 'boot'],
    ) -> Array | None:
        """Point estimate or bootstrap models of the on/off measurement."""
        assert on_off in {'on', 'off'}
        assert mtype in {'mle', 'boot'}

        if (on_off == 'off') and (self.statistic not in _STATISTIC_WITH_BACK):
            return None

        name = f'{self.name}_N{on_off}_model'
        return getattr(self, f'get_model_{mtype}')(name)

    def deviance(self, rtype: Literal['mle', 'boot']) -> Array | None:
        """MLE and bootstrap deviance."""
        if rtype == 'mle':
            return self.result._deviance['point'][self.name]
        elif rtype == 'boot':
            if self.boot is not None:
                return self.boot.deviance['point'][self.name]
            else:
                return None
        else:
            raise ValueError(f'unknown deviance type: {rtype}')

    @property
    def _nsim(self) -> int:
        return 10000

    @_to_cached_method
    def pit(self) -> tuple[Array, Array]:
        stat = self.statistic

        if stat in _STATISTIC_SPEC_NORMAL:
            on_data = self.net_counts
        else:
            on_data = self.spec_counts
        on_model = self.model('on', 'mle')

        if stat in _STATISTIC_SPEC_NORMAL:  # chi2
            pit = stats.norm.cdf((on_data - on_model) / self.net_errors)
            return pit, pit

        if stat in _STATISTIC_WITH_BACK:
            off_data = self.back_counts
            off_model = self.model('off', 'mle')

            if stat in _STATISTIC_BACK_NORMAL:  # pgstat
                pit = pit_poisson_normal(
                    k=on_data,
                    lam=on_model,
                    v=off_data,
                    mu=off_model,
                    sigma=self.back_errors,
                    ratio=self.back_ratio,
                    seed=self.seed + 1,
                    nsim=self._nsim,
                )
                return pit, pit

            else:  # wstat
                return pit_poisson_poisson(
                    k1=on_data,
                    k2=off_data,
                    lam1=on_model,
                    lam2=off_model,
                    ratio=self.data.back_ratio,
                    minus=True,
                    seed=self.seed + 1,
                    nsim=self._nsim,
                )

        else:  # cstat, or pstat
            return pit_poisson(k=on_data, lam=on_model, minus=True)

    def residuals(
        self,
        rtype: Literal['rd', 'rp', 'rq'],
        seed: int | None = None,
        random_quantile: bool = True,
        mle: bool = True,
    ) -> Array | tuple[Array, bool | Array, bool | Array]:
        if rtype == 'rd':
            return self.deviance_residuals_mle()
        elif rtype == 'rp':
            return self.pearson_residuals_mle()
        elif rtype == 'rq':
            seed = self.seed if seed is None else int(seed)
            return self.quantile_residuals_mle(seed, random_quantile)
        else:
            raise NotImplementedError(f'{rtype} residual')

    def residuals_sim(
        self,
        rtype: Literal['rd', 'rp', 'rq'],
        seed: int | None = None,
        random_quantile: bool = True,
    ) -> Array | None:
        if self.boot is None or rtype == 'rq':
            return None

        if rtype == 'rd':
            r = self.deviance_residuals_boot()
        elif rtype == 'rp':
            r = self.pearson_residuals_boot()
        else:
            raise NotImplementedError(f'{rtype} residual')

        return r

    def residuals_ci(
        self,
        rtype: Literal['rd', 'rp', 'rq'],
        cl: float = 0.683,
        seed: int | None = None,
        random_quantile: bool = True,
        with_sign: bool = False,
    ) -> Array | None:
        if self.boot is None or rtype == 'rq':
            return None

        assert 0 < cl < 1

        r = self.residuals_sim(rtype, seed, random_quantile)

        if with_sign:
            return np.quantile(r, q=0.5 + cl * np.array([-0.5, 0.5]), axis=0)
        else:
            q = np.quantile(np.abs(r), q=cl, axis=0)
            return np.row_stack([-q, q])

    @_to_cached_method
    def deviance_residuals_mle(self) -> Array:
        return self._deviance_residuals('mle')

    @_to_cached_method_with_check
    def deviance_residuals_boot(self) -> Array | None:
        return self._deviance_residuals('boot')

    def _deviance_residuals(
        self, rtype: Literal['mle', 'boot']
    ) -> Array | None:
        if rtype == 'boot' and self.boot is None:
            return None

        # NB: if background is present, then this assumes the background is
        #     being profiled out, so that each src & bkg data pair has ~1 dof
        return self.sign[rtype] * np.sqrt(self.deviance(rtype))

    @_to_cached_method
    def pearson_residuals_mle(self) -> Array:
        return self._pearson_residuals('mle')

    @_to_cached_method_with_check
    def pearson_residuals_boot(self) -> Array | None:
        return self._pearson_residuals('boot')

    def _pearson_residuals(
        self, rtype: Literal['mle', 'boot']
    ) -> Array | None:
        if rtype == 'boot' and self.boot is None:
            return None

        stat = self.statistic

        if rtype == 'mle':
            if stat in _STATISTIC_SPEC_NORMAL:
                on_data = self.net_counts
            else:
                on_data = self.spec_counts
        else:
            on_data = self.get_data_boot(f'{self.name}_Non')

        if stat in _STATISTIC_SPEC_NORMAL:
            std = self.net_errors
        else:
            std = None

        r = pearson_residuals(on_data, self.model('on', rtype), std)

        if stat in _STATISTIC_WITH_BACK:
            if rtype == 'mle':
                off_data = self.back_counts
            else:
                off_data = self.get_data_boot(f'{self.name}_Noff')

            if self.statistic in _STATISTIC_BACK_NORMAL:
                std = self.back_errors
            else:
                std = None

            r_b = pearson_residuals(off_data, self.model('off', rtype), std)

            # NB: this assumes the background is being profiled out,
            #     so that each src & bkg data pair has ~1 dof
            r = self.sign[rtype] * np.sqrt(r * r + r_b * r_b)

        return r

    def quantile_residuals_mle(
        self, seed: int, random: bool
    ) -> tuple[Array, Array | bool, Array | bool]:
        pit_minus, pit = self.pit()

        if random:
            pit = np.random.default_rng(seed).uniform(pit_minus, pit)
        r = stats.norm.ppf(pit)

        lower = upper = False

        stat = self.statistic
        if stat == 'chi2':
            mask = (pit == 0.0) | (pit == 1.0)
            if np.any(mask):
                on_data = self.net_counts[mask]
                on_model = self.model('on', 'mle')[mask]
                error = self.net_errors[mask]
                r[mask] = (on_data - on_model) / error

        elif stat in {'cstat', 'pstat'}:
            mask = (pit == 0.0) | (pit == 1.0)
            if np.any(mask):
                on_data = self.spec_counts[mask]
                on_model = self.model('on', 'mle')[mask]
                r[mask] = quantile_residuals_poisson(
                    on_data,
                    on_model,
                    keep_sign=not random,
                    random=random,
                    seed=seed,
                )

        elif stat in {'pgstat', 'wstat'}:
            upper_mask = pit == 0.0
            if np.any(upper_mask):
                r[upper_mask] = stats.norm.ppf(1.0 / self._nsim)
                upper = np.full(r.shape, False)
                upper[upper_mask] = True

            lower_mask = pit == 1.0
            if np.any(lower_mask):
                r[lower_mask] = stats.norm.ppf(1.0 - 1.0 / self._nsim)
                lower = np.full(r.shape, False)
                lower[lower_mask] = True

        return r, lower, upper


# clean up helpers
del (
    _cached_method,
    _cached_method_with_check,
    _to_cached_method,
    _to_cached_method_with_check,
)

_cached_method = []
_cached_method_with_check = []
_to_cached_method = _get_cached_method_decorator(_cached_method)
_to_cached_method_with_check = _get_cached_method_with_check_decorator(
    _cached_method_with_check, 'ppc'
)


class PosteriorPlotData(PlotData):
    result: PosteriorResult
    _cached_method = _cached_method
    _cached_method_with_check = _cached_method_with_check

    @property
    def params(self) -> dict[str, Array]:
        return self.result._params_dist

    @property
    def ppc(self) -> PPCResult | None:
        return self.result._ppc

    @_to_cached_method
    def get_model_median(self, name: str) -> Array:
        posterior = self.result.idata['posterior'][name]
        return posterior.median(dim=('chain', 'draw')).values

    @_to_cached_method
    def get_model_loo(self, name: str) -> Array:
        posterior = self.result.idata['posterior'][name]
        posterior = posterior.stack(__sample__=('chain', 'draw'))
        return self.result._loo_expectation(posterior, self.name).values

    @_to_cached_method
    def get_model_posterior(self, name: str) -> DataArray:
        posterior = self.result.idata['posterior'][name]
        # return shape (n_samples, n_channel)
        return posterior.stack(__sample__=('chain', 'draw')).T

    def get_model_ppc(self, name: str) -> Array | None:
        if self.ppc is None:
            return None
        else:
            return self.ppc.models_fit[name]

    def get_model_mle(self, name: str) -> Array | None:
        mle = self.result._mle
        if mle is None:
            return None
        else:
            return mle['models'][name]

    @property
    def ce_model(self) -> Array:
        return self.get_model_median(self.name)

    @_to_cached_method
    def ce_model_ci(self, cl: float = 0.683) -> Array:
        assert 0.0 < cl < 1.0
        return np.quantile(
            self.get_model_posterior(self.name).values,
            q=0.5 + cl * np.array([-0.5, 0.5]),
            axis=0,
        )

    def unfolded_model(
        self,
        mtype: Literal['ne', 'ene', 'eene'],
        egrid: Array | None,
        params: dict | None,
        comps: bool,
        cl: float | Array | None = None,
    ) -> tuple[Array | dict, Array | dict | None]:
        assert mtype in {'ne', 'ene', 'eene'}
        if cl is not None:
            cl = np.atleast_1d(cl).astype(float)
            assert np.all(0.0 < cl) and np.all(cl < 1.0)
        params = {} if params is None else dict(params)
        comps = comps and self.has_comps
        egrid = jnp.asarray(egrid, float)
        post_params = self.params
        n = [i.size for i in post_params.values()][0]
        params = {k: jnp.full(n, v) for k, v in params.items()}
        post_params = post_params | params
        models = self._unfolded_model(mtype, egrid, post_params, comps)
        if comps:
            model = {k: np.median(v, axis=0) for k, v in models.items()}
            if cl is None:
                return model, None
            q = 0.5 + cl[:, None] * np.array([-0.5, 0.5])
            ci = {k: np.quantile(v, q, axis=0) for k, v in models.items()}
            return model, ci
        else:
            model = np.median(models, axis=0)
            q = 0.5 + cl[:, None] * np.array([-0.5, 0.5])
            ci = np.quantile(models, q, axis=0)
        return model, ci

    @property
    def sign(self) -> dict[str, Array | None]:
        """Sign of the difference between the data and the fitted models."""
        return {
            'posterior': self._sign_posterior(),
            'loo': self._sign_loo(),
            'median': self._sign_median(),
            'mle': self._sign_mle(),
            'ppc': self._sign_ppc(),
        }

    @_to_cached_method
    def _sign_posterior(self) -> Array:
        ce_posterior = self.get_model_posterior(self.name)
        return np.where(self.ce_data >= ce_posterior, 1.0, -1.0)

    @_to_cached_method
    def _sign_loo(self) -> Array:
        ce_loo = self.get_model_loo(self.name)
        return np.where(self.ce_data >= ce_loo, 1.0, -1.0)

    @_to_cached_method
    def _sign_median(self) -> Array:
        ce_median = self.get_model_median(self.name)
        return np.where(self.ce_data >= ce_median, 1.0, -1.0)

    @_to_cached_method_with_check
    def _sign_mle(self) -> Array | None:
        if self.ppc is None:
            return None

        ce_mle = self.get_model_mle(self.name)
        return np.where(self.ce_data >= ce_mle, 1.0, -1.0)

    @_to_cached_method_with_check
    def _sign_ppc(self) -> Array | None:
        if self.ppc is None:
            return None

        ce_ppc = self.get_model_ppc(self.name)
        return np.where(self.ppc.data[self.name] >= ce_ppc, 1.0, -1.0)

    def model(
        self,
        on_off: Literal['on', 'off'],
        mtype: Literal['posterior', 'loo', 'median', 'mle', 'ppc'],
    ) -> Array | None:
        assert on_off in {'on', 'off'}
        assert mtype in {'posterior', 'loo', 'median', 'mle', 'ppc'}

        if (on_off == 'off') and (self.statistic not in _STATISTIC_WITH_BACK):
            return None

        name = f'{self.name}_N{on_off}_model'
        return getattr(self, f'get_model_{mtype}')(name)

    def deviance(
        self,
        rtype: Literal['posterior', 'loo', 'mle', 'ppc'],
    ) -> DataArray | None:
        """Median, MLE, and ppc deviance."""
        if rtype == 'posterior':
            loglike = self.result.idata['log_likelihood'][self.name]
            return -2.0 * loglike.stack(__sample__=('chain', 'draw')).T
        elif rtype == 'loo':
            loglike = self.result.idata['log_likelihood'][self.name]
            deviance = -2.0 * loglike.stack(__sample__=('chain', 'draw')).T
            return self.result._loo_expectation(deviance, self.name)
        elif rtype == 'mle':
            if self.result._mle is not None:
                return self.result._mle['deviance']['point'][self.name]
            else:
                return None
        elif rtype == 'ppc':
            if self.ppc is not None:
                return self.ppc.deviance['point'][self.name]
            else:
                return None
        else:
            raise ValueError(f'unknown deviance type: {rtype}')

    def pit(self) -> tuple:
        return self.result._loo_pit[self.name]

    def residuals(
        self,
        rtype: Literal['rd', 'rp', 'rq'],
        seed: int | None = None,
        random_quantile: bool = True,
        mle: bool = False,
    ) -> Array | tuple[Array, bool | Array, bool | Array]:
        assert rtype in {'rd', 'rp', 'rq'}

        if rtype == 'rq':
            seed = self.seed if seed is None else int(seed)
            return self.quantile_residuals(seed, random_quantile)
        else:
            point_type = 'mle' if mle else 'loo'
            rname = 'deviance' if rtype == 'rd' else 'pearson'
            return getattr(self, f'{rname}_residuals_{point_type}')()

    def residuals_sim(
        self,
        rtype: Literal['rd', 'rp', 'rq'],
        seed: int | None = None,
        random_quantile: bool = True,
    ) -> Array | None:
        if self.ppc is None or rtype == 'rq':
            return None

        if rtype == 'rd':
            r = self.deviance_residuals_ppc()
        elif rtype == 'rp':
            r = self.pearson_residuals_ppc()
        else:
            raise NotImplementedError(f'{rtype} residual')
        return r

    def residuals_ci(
        self,
        rtype: Literal['rd', 'rp', 'rq'],
        cl: float = 0.683,
        seed: int | None = None,
        random_quantile: bool = True,
        with_sign: bool = False,
    ) -> Array | None:
        if self.ppc is None or rtype == 'rq':
            return None

        assert 0 < cl < 1

        r = self.residuals_sim(rtype, seed, random_quantile)

        if with_sign:
            return np.quantile(r, q=0.5 + cl * np.array([-0.5, 0.5]), axis=0)
        else:
            q = np.quantile(np.abs(r), q=cl, axis=0)
            return np.row_stack([-q, q])

    @_to_cached_method
    def deviance_residuals_loo(self) -> Array:
        return self._deviance_residuals('loo')

    @_to_cached_method
    def deviance_residuals_median(self) -> Array:
        return np.median(self._deviance_residuals('posterior'), axis=0)

    @_to_cached_method_with_check
    def deviance_residuals_mle(self) -> Array:
        return self._deviance_residuals('mle')

    @_to_cached_method_with_check
    def deviance_residuals_ppc(self) -> Array | None:
        if self.ppc is None:
            return None
        return self._deviance_residuals('ppc')

    def _deviance_residuals(
        self, rtype: Literal['loo', 'posterior', 'mle', 'ppc']
    ) -> Array | None:
        if rtype in ['mle', 'ppc'] and self.ppc is None:
            return None

        # NB: if background is present, then this assumes the background is
        #     being profiled out, so that each src & bkg data pair has ~1 dof
        return self.sign[rtype] * np.sqrt(self.deviance(rtype))

    @_to_cached_method
    def pearson_residuals_loo(self) -> Array:
        return self._pearson_residuals('loo')

    @_to_cached_method
    def pearson_residuals_median(self) -> Array:
        return np.median(self._pearson_residuals('posterior'), axis=0)

    @_to_cached_method_with_check
    def pearson_residuals_mle(self) -> Array:
        return self._pearson_residuals('mle')

    @_to_cached_method_with_check
    def pearson_residuals_ppc(self) -> Array | None:
        if self.ppc is None:
            return None
        return self._pearson_residuals('ppc')

    def _pearson_residuals(
        self, rtype: Literal['posterior', 'loo', 'mle', 'ppc']
    ) -> Array | None:
        if rtype in ['mle', 'ppc'] and self.ppc is None:
            return None

        stat = self.statistic
        mtype = 'posterior' if rtype == 'loo' else rtype

        if rtype in {'posterior', 'loo', 'mle'}:
            if stat in _STATISTIC_SPEC_NORMAL:
                on_data = self.net_counts
            else:
                on_data = self.spec_counts
        else:
            on_data = self.ppc.data[f'{self.name}_Non']
        on_model = self.model('on', mtype)

        if stat in _STATISTIC_SPEC_NORMAL:
            std = self.net_errors
        else:
            std = None

        r = pearson_residuals(on_data, on_model, std)

        if stat in _STATISTIC_WITH_BACK:
            if rtype in {'posterior', 'loo', 'mle'}:
                off_data = self.back_counts
            else:
                off_data = self.ppc.data[f'{self.name}_Noff']
            off_model = self.model('off', mtype)

            if stat in _STATISTIC_BACK_NORMAL:
                std = self.back_errors
            else:
                std = None

            r_b = pearson_residuals(off_data, off_model, std)

            # NB: this assumes the background is being profiled out,
            #     so that each src & bkg data pair has ~1 dof
            r = self.sign[rtype] * np.sqrt(r * r + r_b * r_b)

        if rtype == 'loo':
            r = self.result._loo_expectation(np.abs(r), self.name)
            r *= self.sign[rtype]

        return r

    def quantile_residuals(
        self, seed: int, random: bool
    ) -> tuple[Array, Array | bool, Array | bool]:
        pit_minus, pit = self.pit()
        if random:
            pit = np.random.default_rng(seed).uniform(pit_minus, pit)
        r = stats.norm.ppf(pit)

        # Assume the posterior prediction is nchan * ndraw times
        nchain = len(self.result.idata['posterior']['chain'])
        ndraw = len(self.result.idata['posterior']['draw'])
        nsim = nchain * ndraw

        lower = upper = False

        upper_mask = pit == 0.0
        if np.any(upper_mask):
            r[upper_mask] = stats.norm.ppf(1.0 / nsim)
            upper = np.full(r.shape, False)
            upper[upper_mask] = True

        lower_mask = pit == 1.0
        if np.any(lower_mask):
            r[lower_mask] = stats.norm.ppf(1.0 - 1.0 / nsim)
            lower = np.full(r.shape, False)
            lower[lower_mask] = True

        return r, lower, upper


# clean up helpers
del (
    _cached_method,
    _cached_method_with_check,
    _to_cached_method,
    _to_cached_method_with_check,
)



from __future__ import annotations

from itertools import permutations
from typing import TYPE_CHECKING

import numpy as np
import zeus
import zeus.ensemble as zeus_ensemble
import zeus.moves as zeus_moves

from elisa.infer.samplers.ensemble.base import (
    EnsembleSampler,
    EnsembleSamplerState,
)

if TYPE_CHECKING:
    from multiprocessing import Queue

    from numpy.typing import NDArray


class ZeusSampler(EnsembleSampler):
    def get_sampling_fn(
        self,
        chains: int,
        warmup: int,
        steps: int,
        thinning: int,
        tune: bool | None,
        warmup_kwargs: dict,
        sampling_kwargs: dict,
    ):
        ndim = self._ndim
        if tune is None:
            tune = True
        warmup_kwargs.setdefault('verbose', False)
        sampling_kwargs.setdefault('verbose', False)
        log_prob_fn = self._log_prob_with_blobs
        blobs_dtype = self._blobs_dtype

        def sampling_fn(
            sampler_id: int,
            state: EnsembleSamplerState,
            queue: Queue,
        ):
            np.random.set_state(state.random_state)
            start = state.coords
            sampler1 = zeus.EnsembleSampler(
                chains,
                ndim,
                log_prob_fn,
                args=None,
                kwargs=None,
                tune=tune,
                pool=None,
                vectorize=True,
                blobs_dtype=blobs_dtype,
                **warmup_kwargs,
            )
            sampler2 = zeus.EnsembleSampler(
                chains,
                ndim,
                log_prob_fn,
                args=None,
                kwargs=None,
                tune=tune,
                pool=None,
                vectorize=True,
                blobs_dtype=blobs_dtype,
                **sampling_kwargs,
            )
            queue.put((sampler_id, 'warmup'))
            for _ in sampler1.sample(
                start=start,
                iterations=warmup,
                progress=False,
            ):
                queue.put((sampler_id, 'update'))
            if warmup:
                start = sampler1.get_last_sample()
            sampler1.reset()
            queue.put((sampler_id, 'sample'))
            for _ in sampler2.sample(
                start=start,
                iterations=steps,
                thin_by=thinning,
                progress=False,
            ):
                queue.put((sampler_id, 'update'))
            queue.put((sampler_id, 'finish'))
            samples = sampler2.get_blobs()
            state = EnsembleSamplerState(
                coords=sampler2.get_last_sample(),
                random_state=np.random.get_state(legacy=True),
            )
            return samples, state

        return sampling_fn

    def get_random_state(self, seed: int) -> tuple:
        return np.random.RandomState(seed).get_state(legacy=True)


class DifferentialMove:
    """Improved DifferentialMove of zeus for reproducibility."""

    def __init__(self, tune=True, mu0=1.0):
        self.tune: bool = tune
        self.mu0: float = mu0
        self._nsamples: int | None = None
        self._perms: NDArray[int] | None = None
        self._nperms: int | None = None

    def get_direction(self, X: NDArray[float], mu: float):
        nsamples = X.shape[0]
        if nsamples != self._nsamples:
            self._nsamples = nsamples
            self._perms = np.array(list(permutations(np.arange(nsamples), 2)))
            self._nperms = len(self._perms)

        idx = np.random.choice(self._nperms, self._nsamples, replace=False)
        pairs = self._perms[idx].T

        if not self.tune:
            mu = self.mu0

        return 2.0 * mu * (X[pairs[0]] - X[pairs[1]]), self.tune


# Monkey patching the DifferentialMove for reproducibility
zeus_ensemble.DifferentialMove = DifferentialMove
zeus_moves.DifferentialMove = DifferentialMove



"""Bi-Symmetric log transformation for uniform distribution.

See https://iopscience.iop.org/article/10.1088/0957-0233/24/2/027001.

Contributed by @xiesl97 (https://github.com/xiesl97).
"""

from __future__ import annotations

import jax
import jax.numpy as jnp
from numpyro.distributions import Uniform, constraints
from numpyro.distributions.distribution import (
    Transform,
    TransformedDistribution,
)
from numpyro.distributions.util import promote_shapes


@jax.jit
def log(x, c=0.43429448190325176):
    """
    transformation  x -> y
    c default to 1/log(10)
    """
    return jnp.sign(x) * jnp.log10(1 + jnp.abs(x / c))


@jax.jit
def pow(y, c=0.43429448190325176):
    """
    inverse transformation  y -> x
    c default to 1/log(10)
    """
    return jnp.sign(y) * c * (-1 + jnp.power(10, jnp.abs(y)))


class BiSymTransform(Transform):
    # TODO: refine domain/codomain Bi-Symmetric logic through setters,
    #  especially when transforms for inverses are supported
    def __init__(self, domain=constraints.real, c=0.43429448190325176):
        self.domain = domain
        self._c = c

    @property
    def codomain(self):
        if self.domain is constraints.ordered_vector:
            return constraints.positive_ordered_vector
        elif self.domain is constraints.real:
            return constraints.positive
        elif isinstance(self.domain, constraints.greater_than):
            return constraints.greater_than(
                self.__call__(self.domain.lower_bound)
            )
        elif isinstance(self.domain, constraints.interval):
            return constraints.interval(
                self.__call__(self.domain.lower_bound),
                self.__call__(self.domain.upper_bound),
            )
        else:
            raise NotImplementedError

    def __call__(self, x):
        # XXX consider to clamp from below for stability if necessary
        return pow(x, self._c)

    def _inverse(self, y):
        return log(y, self._c)

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        return jnp.log(jnp.pow(10.0, jnp.abs(x)))

    def tree_flatten(self):
        return (self.domain,), (('domain',), {})

    def __eq__(self, other):
        if not isinstance(other, BiSymTransform):
            return False
        return self.domain == other.domain


class BiSymLogUniform(TransformedDistribution):
    """
    the input should be transfor before input

    for example:
    import numpy as np
    import bslogu as bs
    from numpyro.distributions import LogUniform, Uniform

    Uniform( low , high )
    LogUniform( np.log(low) , np.log(high) )
    BiSymLogUniform( bs.log(low) , bs.log(high) )
    """

    arg_constraints = {
        'low': constraints.dependent,
        'high': constraints.dependent,
    }
    reparametrized_params = ['low', 'high']
    pytree_data_fields = ('low', 'high', '_support')

    def __init__(
        self,
        low,
        high,
        *,
        c=0.43429448190325176,
        validate_args=None,
    ):
        base_dist = Uniform(log(low, c), log(high, c))
        self._c = c
        self.low, self.high = promote_shapes(low, high)
        self._support = constraints.interval(self.low, self.high)
        super().__init__(
            base_dist, BiSymTransform(c=c), validate_args=validate_args
        )

    @constraints.dependent_property(is_discrete=False, event_dim=0)
    def support(self):
        return self._support

    @property
    def mean(self):
        raise NotImplementedError

    @property
    def variance(self):
        raise NotImplementedError

    def cdf(self, x):
        return self.base_dist.cdf(log(x, self._c))
